# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Utilities module within the ketos library

    This module provides a number of auxiliary methods.
"""

import os
import numpy as np
import pandas as pd
from scipy.signal import find_peaks
from functools import reduce

def factors(n):    
    """ Returns sorted set of all divisors of n

        Args:
            n: int
                Integer number

        Returns:
            s: set
                Sorted set of all divisors of n
    """
    s = set(reduce(list.__add__, 
                ([i, n//i] for i in range(1, int(n**0.5) + 1) if n % i == 0)))
    return s

def ensure_dir(file_path):
    """ Ensure that destination directory exists.

        If the directory does not exist, it is created.
        If it already exists, nothing happens.
        
        Args:
            file_path: str
                Full path to destination
    """
    directory = os.path.dirname(file_path)
    if not os.path.exists(directory):
        os.makedirs(directory)

def random_floats(size=1, low=0, high=1, seed=1):
    """ Returns a random number or numpy array of randum numbers uniformly distributed in the half-open interval [low, high)
        
        Args:
            size: int
                Number of random numbers to be generated
            low: float
                Lower value
            high: float
                Upper value (not included)
            seed: int
                Seed for the random number generator

        Returns:
            res: float or numpy array
                Generated random number(s)

        Example:
            >>> from ketos.utils import random_floats
            >>> 
            >>> x = random_floats(3, 0.4, 7.2)
            >>> print(x)
            [3.23574963 5.29820656 0.40077775]
    """

    assert high >= low, "Upper limit must be greater than or equal to lower limit"
    assert size >= 1, "Size must be an int greater than or equal to 1"

    np.random.seed(seed)

    if high == low:
        if size == 1:
            res = high
        else:
            res = np.ones(size, dtype=float) * high
    
    else:
        rndm = np.random.random_sample(size)
        res = low + (high - low) * rndm
        if len(res) == 1:
            res = np.float(res)

    return res

def ndim(a):
    """ Returns the number of dimensions of a list/tuple/array.
        
        Args:
            a: list, tuple or numpy array
                Object that we wish to obtain the dimension of 

        Returns:
            n: int
                Number of dimensions

        Example:
            >>> from ketos.utils import ndim
            >>> 
            >>> x = [[0, 1, 2, 3],(4, 5)]
            >>> print(ndim(x))
            2
    """
    if not (type(a) == list or type(a) == tuple or type(a) == np.ndarray):
        return 0

    if len(a) == 0:
        return 1

    n = 1 + ndim(a[0])
    return n

def tostring(box, decimals=None):
    """ Convert an array, tuple or list into a string.

        Args:
            box: array, tuple or list
                Array, tuple or list that will be converted into a string.
            decimals: int
                Number of decimals that will be kept in the conversion to string.

        Returns:
            s: str
                String representation of array/tuple/list.

        Example:
            >>> from ketos.utils import tostring
            >>> 
            >>> y = [[0, 1, 2, 3],(4, 5)]
            >>> print(tostring(y))
            [[0,1,2,3],[4,5]]
    """
    if box is None:
        return ''

    box = np.array(box)

    if decimals is not None:
        box = np.around(box, decimals=int(decimals))

    box = box.tolist()

    s = str(box)
    s = s.replace(' ', '')
    s = s.replace('(', '[')
    s = s.replace(')', ']')

    return s

def octave_bands(band_min=-1, band_max=9):
    """ Compute the min, central, and max frequency value 
        of the specified octave bands, using the following 
        formulas,

            f_centre = 10^3 * 2^p ,
            f_min = f_centre / sqrt(2) ,
            f_max = f_centre * sqrt(2) ,

        where p = band_number - 5

        Args:
            band_min: int
                Lower octave band
            band_max: int
                Upper octave band

        Returns:
            fcentre: numpy array
                Central frequency of each band (in Hz)
            flow: numpy array
                Minimum frequency of each band (in Hz)
            fhigh: numpy array
                Maximum frequency of each band (in Hz)

        Example:
            >>> from ketos.utils import octave_bands
            >>> 
            >>> fc, fmin, fmax = octave_bands(1, 3)
            >>> print(fc)
            [ 62.5 125.  250. ]
    """
    p = np.arange(band_min-5., band_max-4.)
    fcentre = np.power(10.,3) * np.power(2.,p)
    fd = np.sqrt(2.)
    flow = fcentre / fd
    fhigh = fcentre * fd
    return fcentre, flow, fhigh

def octave_bands_json(band_min, band_max):
    """ Produce a string of the specified octave bands
        in json format

        Args:
            band_min: int
                Lower octave band
            band_max: int
                Upper octave band

        Returns:
            s: str
                json format string

        Example:
            >>> from ketos.utils import octave_bands_json
            >>> 
            >>> s = octave_bands_json(1, 2)
    """
    fcentre, flow, fhigh = octave_bands(band_min, band_max)

    s = "\"frequency_bands\": [\n"
    n = len(flow)
    for i in range(n):
        s += "\t{\n"
        s += "\t\t\"name\": \"{0:.0f}Hz\",\n".format(fcentre[i])
        s += "\t\t\"range\": [\"{0:.1f}Hz\", \"{1:.1f}Hz\"]".format(flow[i],fhigh[i])
        endpar = "\n\t}"
        if i < n-1:
            endpar += ","

        s += endpar + "\n"

    s += "]"
    return s

def morlet_func(time, frequency, width, displacement, norm=True, dfdt=0):
    """ Compute Morlet wavelet function

        The function is implemented as in Eq. (15) in John Ashmead, "Morlet Wavelets in Quantum Mechanics",
        Quanta 2012; 1: 58-70, with the replacement f -> 2*pi*f*s, to allow f to be identified with the 
        physical frequency.

        Args:
            time: float or numpy array
               Time in seconds at which the function is to be evaluated
            frequency: float
                Wavelet frequency in Hz
            width: float
                Wavelet width in seconds (1-sigma width of the Gaussian envelope function)
            displacement: float
                Wavelet centroid in seconds
            norm: bool
                Include [pi^1/4*sqrt(sigma)]^-1 normalization factor
            dfdt: float
                Rate of change in frequency as a function of time in Hz per second.
                If dfdt is non-zero, the frequency is computed as 
                    
                    f = frequency + (time - displacement) * dfdt 

        Returns:
            y: float or numpy array
                Value of Morlet wavelet function at time t

        Example:
            >>> from ketos.utils import morlet_func
            >>> 
            >>> time = np.array([-1., 0., 0.5])
            >>> f = morlet_func(time=time, frequency=10, width=3, displacement=0)
            >>> print(f)
            [0.41022718 0.43366254 0.42768108]
    """
    if dfdt != 0:
        frequency += (time - displacement) * dfdt
    
    assert np.all(frequency > 0), "Frequency must be a strictly positive float"
    assert width > 0, "Width must be a strictly positive float"

    t = time
    w = 2 * np.pi * frequency * width
    s = width
    l = displacement
    x = (t-l)/s

    y = (np.exp(1j*w*x) - np.exp(-0.5*(w**2))) * np.exp(-0.5*(x**2))

    if norm:
        y *= (s * np.sqrt(np.pi) * (1 + np.exp(-w**2) - 2*np.exp(-0.75*w**2)) )**-0.5

    return np.real(y)


def nearest_values(x, i, n):
    """ Returns the n values nearest to index i from the array x.

        Here, nearest refers to the position in the array, not 
        the value.

        Args:
            x: numpy array
                Input values
            i: int
                Index
            n: int
                Number of neighboring values

        Returns:
            y: numpy array
                n values nearest to index i from the array x 

        Example:
            >>> from ketos.utils import nearest_values
            >>> 
            >>> x = np.array([1.0, 4.0, 5.1, 6.0, 0.2, 0.3, 10.0])
            >>> y = nearest_values(x=x, i=3, n=3)
            >>> print(y)
            [5.1 6.  0.2]
    """
    if n >= x.shape[0]:
        return x

    if n%2 == 0:
        i1 = int(i - n/2)
        i2 = int(i + n/2 - 1)
    else:
        i1 = int(i - (n-1)/2)
        i2 = int(i + (n-1)/2)

    if i1 >= 0 and i2 < x.shape[0]:
        return x[i1:i2+1]

    v = list()
    v.append(x[i])
    k = 1
    while len(v) < n:
        if k%2 == 0:
            d = k/2
        else:
            d = (k+1)/2
        j = int(i + np.power(-1,k) * d)
        if j >= 0 and j < x.shape[0]:
            v.append(x[j])
        k += 1
    v = np.array(v)
    return v


def detect_peaks(df, distance=1, multiplicity=1, prominence=1.0, height=None, threshold=None):
    """ Detect peaks in time-series data.

        The time-series data is provided in the form of a Pandas DataFrame object, where 
        each column contains a different time series.

        This is essentially a wrapper around a SciPy's find_peaks method:

            https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.find_peaks.html

        Args:
            df: Pandas DataFrame
                Data frame containing the input data.
            distance: int
                Minimum distance between adjacent peaks
            multiplicity: int
                Number of time series in which peaks must appear to be counted.
            prominence: float
                Required prominence of the peaks. The prominence of a peak measures how much a peak 
                stands out from the surrounding baseline of the signal and is defined as the vertical 
                distance between the peak and its lowest contour line. See also 
                https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.peak_prominences.html#scipy.signal.peak_prominences
            height: float
                Required absolute height of the peaks.
            threshold: float
                Required threshold of peaks (the vertical distance to its neighbouring samples).

        Returns:
            y: Pandas DataFrame 
                Data frame containing the detected peaks

        Example:
            >>> from ketos.utils import detect_peaks
            >>> import pandas as pd
            >>>
            >>> # create a two time series, where only the first contains a peak
            >>> d = {'series1' : pd.Series([1.0, 2.3, 22.0, 2.2, 1.5]), 'series2': pd.Series([1.0, 2.3, 1.8, 2.2, 1.5])}
            >>> df = pd.DataFrame(d)
            >>> 
            >>> # detect peaks with multiplicity 1 and prominence of at least 2.0
            >>> peaks = detect_peaks(df=df, multiplicity=1, prominence=2.0)
            >>> print(peaks)
            [0 0 1 0 0]
            >>> 
            >>> # try again, but this time require multiplicity 2
            >>> peaks = detect_peaks(df=df, multiplicity=2, prominence=2.0)
            >>> print(peaks)
            [0 0 0 0 0]
    """
    peaks = pd.DataFrame(index=df.index)

    for column in df:
        x = df[column]
        m = np.median(np.abs(x - np.median(x)))
        if prominence > 0:
            min_prominence = m * prominence
        else:
            min_prominence = None
        positions, _ = find_peaks(x, height=height, threshold=threshold, distance=distance, prominence=(min_prominence,None))
        y = np.zeros(len(x))
        y[positions] = 1
        peaks[column] = y

    # sum across each row
    peaks = peaks.sum(axis=1) 

    # create column of FALSEs (no detection) and TRUES (detection)
    detections = pd.DataFrame((peaks >= multiplicity)) 

    # convert to 0s and 1s and extract numpy array
    res = detections[detections.columns[0]].astype(int).values 

    return res


def get_member(cls, member_name):
    """ Query class member by name.

        Returns ValueError if class does not contain a member by that name.

        Args:
            cls: Class
                Class
            member_name: str
                Member name

        Returns:
            member: 
                The class member
    """
    for name, member in cls.__members__.items():
        if member_name == name:
            return member

    s = ", ".join(name for name, _ in cls.__members__.items())
    raise ValueError("Unknown value \'{0}\'. Select between: {1}".format(member_name, s))



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Data handling module within the ketos library

    This module provides utilities to load and handle data files.
"""
import numpy as np
import pandas as pd
import librosa
import os
import math
import errno
import tables
from subprocess import call
import scipy.io.wavfile as wave
import ketos.external.wavfile as wave_bit
from ketos.utils import tostring
import datetime
import datetime_glob
import re
from ketos.data_handling.parsing import SpectrogramConfiguration


def rel_path_unix(path, start=None):
    """ Return a relative unix filepath to path either from the current 
        directory or from an optional start directory.

        Args:
            path: str
                Path. Can be unix or windows format.
            start: str
                Optional start directory. Can be unix or windows format.

        Returns:
            u: str
                Relative unix filepath

        Examples:
            >>> from ketos.data_handling.data_handling import rel_path_unix      
            >>> path = "/home/me/documents/projectX/file1.pdf"
            >>> start = "/home/me/documents/"
            >>> u = rel_path_unix(path, start)
            >>> print(u)
            /projectX/
    """
    rel = os.path.relpath(path, start)
    h,t = os.path.split(rel)
    u = '/'
    while len(h) > 0:
        h,t = os.path.split(h)
        u = '/' + t + u

    return u



def parse_datetime(to_parse, fmt=None, replace_spaces='0'):
    """Parse date-time data from string.
       
       Returns None if parsing fails.
        
        Args:
            to_parse: str
                String with date-time data to parse.
            fmt: str
                String defining the date-time format. 
                Example: %d_%m_%Y* would capture "14_3_1999.txt"
                See https://pypi.org/project/datetime-glob/ for a list of valid directives
                
            replace_spaces: str
                If string contains spaces, replaces them with this string

        Returns:
            datetime: datetime object

        Examples:
            >>> #This will parse dates in the day/month/year format,
            >>> #separated by '/'. It will also ignore any text after the year,
            >>> # (such as a file extension )
            >>>
            >>> from ketos.data_handling.data_handling import parse_datetime           
            >>> fmt = "%d/%m/%Y*"
            >>> result = parse_datetime("10/03/1942.txt", fmt)
            >>> result.year
            1942
            >>> result.month
            3
            >>> result.day
            10
            >>>
            >>> # Now with the time (hour:minute:second) separated from the date by an underscore
            >>> fmt = "%H:%M:%S_%d/%m/%Y*"
            >>> result = parse_datetime("15:43:03_10/03/1918.wav", fmt)
            >>> result.year
            1918
            >>> result.month
            3
            >>> result.day
            10
            >>> result.hour
            15
            >>> result.minute
            43
            >>> result.second
            3
    """

    # replace spaces
    to_parse = to_parse.replace(' ', replace_spaces)
    
    if fmt is not None:
        matcher = datetime_glob.Matcher(pattern=fmt)
        match = matcher.match(path=to_parse)
        if match is None:
            return None
        else:
            return match.as_datetime()

    return None

def find_files(path, substr, fullpath=True, subdirs=False):
    """ Find all files in the specified directory containing the specified substring in their file name

        Args:
            path: str
                Directory path
            substr: str
                Substring contained in file name
            fullpath: bool
                If True, return relative path to each file. If false, only return the file names 
            subdirs: bool
                If True, search all subdirectories

        Returns:
            files: list (str)
                Alphabetically sorted list of file names

        Examples:
            >>> from ketos.data_handling.data_handling import find_files
            >>>
            >>> # Find files that contain 'super' in the name;
            >>> # Do not return the relative path
            >>> find_files(path="ketos/tests/assets", substr="super", fullpath=False)
            ['super_short_1.wav', 'super_short_2.wav']
            >>>
            >>> # find all files with '.h5" in the name
            >>> # Return the relative path
            >>> find_files(path="ketos/tests/assets", substr=".h5")
            ['ketos/tests/assets/15x_same_spec.h5', 'ketos/tests/assets/cod.h5', 'ketos/tests/assets/humpback.h5', 'ketos/tests/assets/morlet.h5']
    """
    # find all files
    allfiles = list()
    if not subdirs:
        f = os.listdir(path)
        for fil in f:
            if fullpath:
                x = path
                allfiles.append(os.path.join(x, fil))
            else:
                allfiles.append(fil)
    else:
        for r, _, f in os.walk(path):
            for fil in f:
                if fullpath:
                    allfiles.append(os.path.join(r, fil))
                else:
                    allfiles.append(fil)

    # select those that contain specified substring
    files = list()
    for f in allfiles:
        if substr in f:
            files.append(f)

    # sort alphabetically
    files.sort()

    return files


def find_wave_files(path, fullpath=True, subdirs=False):
    """ Find all wave files in the specified directory

        Args:
            path: str
                Directory path
            fullpath: bool
                Return relative path to each file or just the file name 

        Returns:
            wavefiles: list (str)
                Alphabetically sorted list of file names

        Examples:
            >>> from ketos.data_handling.data_handling import find_wave_files
            >>>
            >>> find_wave_files(path="ketos/tests/assets", fullpath=False)
            ['2min.wav', 'empty.wav', 'grunt1.wav', 'super_short_1.wav', 'super_short_2.wav']

    """
    wavefiles = find_files(path, '.wav', fullpath, subdirs)
    wavefiles += find_files(path, '.WAV', fullpath, subdirs)
    return wavefiles


def read_wave(file, channel=0):
    """ Read a wave file in either mono or stereo mode

        Args:
            file: str
                path to the wave file
            channel: int
                Which channel should be used in case of stereo data (0: left, 1: right) 

        Returns: (rate,data)
            rate: int
                The sampling rate
            data: numpy.array (float)
                A 1d array containing the audio data
        
        Examples:
            >>> from ketos.data_handling.data_handling import read_wave
            >>> rate, data = read_wave("ketos/tests/assets/2min.wav")
            >>> # the function returns the sampling rate (in Hz) as an integer
            >>> type(rate)
            <class 'int'>
            >>> rate
            2000
            >>> # And the actual audio data is a numpy array
            >>> type(data)
            <class 'numpy.ndarray'>
            >>> len(data)
            241664
            >>> # Since each item in the vector is one sample,
            >>> # The duration of the audio in seconds can be obtained by
            >>> # dividing the the vector length by the sampling rate
            >>> len(data)/rate
            120.832
    """
    try:
        rate, signal, _ = wave_bit.read(file)
    except TypeError:
        rate, signal = wave.read(file)
           
    if len(signal.shape) == 2:
        data = signal[:, channel]
    else:
        data = signal[:]
    return rate, data

def create_dir(dir):
    """ Create a new directory if it does not exist

        Will also create any intermediate directories that do not exist
        Args:
            dir: str
               The path to the new directory
     """
    os.makedirs(dir, exist_ok=True)

def to1hot(value,depth):
    """Converts the binary label to one hot format

            Args:
                value: scalar or numpy.array | int or float
                    The the label to be converted.
                depth: int
                    The number of possible values for the labels 
                    (number of categories).
                                
            Returns:
                one_hot:numpy array (dtype=float64)
                    A len(value) by depth array containg the one hot encoding
                    for the given value(s).

            Example:
                >>> from ketos.data_handling.data_handling import to1hot
                >>>
                >>> # An example with two possible labels (0 or 1)
                >>> values = np.array([0,1])
                >>> to1hot(values,depth=2)
                array([[1., 0.],
                       [0., 1.]])
                >>>
                >>> # The same example with 4 possible labels (0,1,2 or 3)
                >>> values = np.array([0,1])
                >>> to1hot(values,depth=4)
                array([[1., 0., 0., 0.],
                       [0., 1., 0., 0.]])
     """
    value = np.int64(value)
    one_hot = np.eye(depth)[value]
    return one_hot

def from1hot(value):
    """Converts the one hot label to binary format

            Args:
                value: scalar or numpy.array | int or float
                    The  label to be converted.
            
            Returns:
                output: int or numpy array (dtype=int64)
                    An int representing the category if 'value' has 1 dimension or an
                    array of m ints if values is an n by m array.

            Example:
                >>> from ketos.data_handling.data_handling import from1hot
                >>>
                >>> from1hot(np.array([0,0,0,1,0]))
                3
                >>> from1hot(np.array([[0,0,0,1,0],
                ...   [0,1,0,0,0]]))
                array([3, 1])

     """

    if value.ndim > 1:
        output = np.apply_along_axis(arr=value, axis=1, func1d=np.argmax)
        output.dtype = np.int64
    else:
        output = np.argmax(value)

    return output


def check_data_sanity(images, labels):
    """ Check that all images have same size, all labels have values, 
        and number of images and labels match.
     
        Args:
            images: numpy array or pandas series
                Images
            labels: numpy array or pandas series
                Labels
        Raises:
            ValueError:
                If no images or labels are passed;
                If the number of images and labels is different;
                If images have different shapes;
                If any labels are NaN.

       Returns:
            True if all checks pass.

        Examples:
            >>> from ketos.data_handling.data_handling import check_data_sanity
            >>> # Load a database with images and integer labels
            >>> data = pd.read_pickle("ketos/tests/assets/pd_img_db.pickle")
            >>> images = data['image']
            >>> labels = data['label']
            >>> # When all the images and labels  pass all the quality checks,
            >>> # The function returns True            
            >>> check_data_sanity(images, labels)
            True
            >>> # If something is wrong, like if the number of labels
            >>> # is different from the number of images, and exeption is raised
            >>> labels = data['label'][:10] 
            >>> check_data_sanity(images, labels=labels)
            Traceback (most recent call last):
                File "/usr/lib/python3.6/doctest.py", line 1330, in __run
                    compileflags, 1), test.globs)
                File "<doctest data_handling.check_data_sanity[5]>", line 1, in <module>
                    check_data_sanity(images, labels=labels)
                File "ketos/data_handling/data_handling.py", line 599, in check_data_sanity
                    raise ValueError("Image and label columns have different lengths")
            ValueError: Image and label columns have different lengths
    """
    checks = True
    if images is None or labels is None:
        raise ValueError(" Images and labels cannot be None")
        

    # check that number of images matches numbers of labels
    if len(images) != len(labels):
        raise ValueError("Image and label columns have different lengths")

    # determine image size and check that all images have same size
    image_shape = images[0].shape
    if not all(x.shape == image_shape for x in images):
        raise ValueError("Images do not all have the same size")

    # check that all labels have values
    b = np.isnan(labels)    
    n = np.count_nonzero(b)
    if n != 0:
        raise ValueError("Some labels are NaN")
    
    return checks

def get_image_size(images):
    """ Get image size and check that all images have same size.
     
        Args:
            images: numpy array or pandas series
                Images

        Results:
            image_size: tuple (int,int)
                Image size

        Examples:
            >>> from ketos.data_handling.data_handling import get_image_size
            >>> import pandas as pd
            >>>
            >>> # Load a dataset with images and integer labels
            >>> data = pd.read_pickle("ketos/tests/assets/pd_img_db.pickle")
            >>>
            >>> # Select only the images from the dataset
            >>> images = data['image']
            >>> get_image_size(images)
            (20, 20)

    """
    # determine image size and check that all images have same size
    image_shape = images[0].shape
    assert all(x.shape == image_shape for x in images), "Images do not all have the same size"

    return image_shape


def parse_seg_name(seg_name):
    """ Retrieves the segment id and label from the segment name

        Args:
            seg_name: str
            Name of the segment in the format id_*_*_l_*.wav, where 'id' is 
            followed by base name of the audio file from which the segment was extracted, '_',
            and a sequence number. The 'l' is followed by any number of characters describing the label(s).

        Raise:
            ValueError
            If seg_name is empty or in wrong format
        Returns:
            (id,label) : tuple (str,str)
            A tuple with the id and label strings.

        Examples:
            >>> from ketos.data_handling.data_handling import parse_seg_name
            >>> seg_name = "id_hydr06_23_l_[2].wav"
            >>> id, label = parse_seg_name(seg_name)
            >>> id
            'hydr06_23'
            >>> label
            '[2]'
            >>>
            >>> seg_name = "id_hydr05_279_l_[2,1].mp3"
            >>> id, label = parse_seg_name(seg_name)
            >>> id
            'hydr05_279'
            >>> label
            '[2,1]'

    """
    id, labels = None, None
    pattern=re.compile('id_(.+)_(.+)_l_\[(.+)\].*')
    if not pattern.match(seg_name):
       raise ValueError("seg_name must follow the format  id_*_*_l_[*].")


    splits = seg_name.split("_")
    if len(splits) >= 5:
        id = seg_name.split("_")[1] + "_" + seg_name.split("_")[2]
        tmp = seg_name.split("_")[4]
        labels = tmp.split(".")[0]
    
    return (id,labels)



def divide_audio_into_segs(audio_file, seg_duration, save_to, annotations=None, start_seg=None, end_seg=None, verbose=False):
    """ Divide a large .wav file into a sequence of smaller segments with the same duration.
        

        Name the resulting segments sequentially and save them as .wav files in the specified directory.
        If annotations are provided, this function will check if the segment created emcompasses any labels. If so,
        the label information will be added to the segment name.
        
        Note: segments will be saved following the name pattern "id_*_*_l_*.wav",
            where 'id_' is followed by the name of the original file, underscore ('_') 
            and the a sequence name. 'l_' is followed by the label(s) associated with that segment.
            Ex: 'id_rec03_87_l_[1,3]', 'id_rec03_88_l_[0]

            The start_seg and end_seg arguments can be used to segment only part of audio files,
            which is useful when processing large files in parallel.
            
        Args:
            audio_file:str
            .wav file name (including path).

            seg_duration: float
            desired duration for each segment

            annotations: pandas.DataFrame
            DataFrame with the the annotations. At least the following columns are expected:
                "filename": the file name. Must be the the same as audio_file
                "label": the label value for each annotaded event
                "start": the start time relative to the beginning of the audio_file.
                "end": the end time relative to the beginning of the file. 
            If None, the segments will be created and file names will have 'NULL' as labels. 
            Ex: 'id_rec03_87_l[NULL].wav.
                    
            save_to: str
            path to the directory where segments will be saved.

            start_seg: int
                Indicates the number of the segment on which the segmentation will start.
                A value of 3 would indicate the 3rd segment in a sequence(if 'seg_duration' is set to 2.0,
                that would corresponfd to 6.0 seconds from the beginning of the file')
            end_seg:int
                Indicates the number of the segment where the segmentation will stop.
                A value of 6 would indicate the 3rd segment in a sequence(if 'seg_duration' is set to 2.0,
                that would correspond to 12.0 seconds from the beginning of the file'
            verbose:bool
                If True, print "Creating segment .... name_of_segment" for each segment.
                        
         Returns:
            None   

        Examples:
            >>> from ketos.data_handling.data_handling import divide_audio_into_segs
            >>> from glob import glob
            >>> import os
            >>>
            >>> # Define the paths to the audio file that will be segmented
            >>> # And the folder where the segments will be saved
            >>> audio_file = "ketos/tests/assets/2min.wav"
            >>> save_dir = "ketos/tests/assets/tmp/divided_segs"
            >>> # Create that folder (if it does not exist)
            >>> os.makedirs(save_dir, exist_ok=True)
            >>> # Difine an annotations dataframe
            >>> annotations = pd.DataFrame({'filename':['2min.wav','2min.wav','2min.wav'],
            ...                    'label':[1,2,1], 'start':[5.0, 70.34, 105.8],
            ...                    'end':[6.0,75.98,110.0]})
            >>>
            >>> # Devide the wav file into 2 seconds segments.
            >>> # Uses the annotations dataframe to determine if each segment
            >>> # includes a label names the segments accordingly
            >>> divide_audio_into_segs(audio_file=audio_file,
            ... seg_duration=2.0, annotations=annotations, save_to=save_dir)
            >>> # Count all files have been created in the destination folder
            >>> n_seg = len(glob(save_dir + "/id_2min*.wav"))
            >>> #60 files have been created
            >>> n_seg
            60
    """
    create_dir(save_to)
    orig_audio_duration = librosa.get_duration(filename=audio_file)
    n_seg = round(orig_audio_duration/seg_duration)

    prefix = os.path.basename(audio_file).split(".wav")[0]

    if start_seg is None:
        start_seg = 0
    if end_seg is None:
        end_seg = n_seg - 1

    for s in range(start_seg, end_seg + 1):
        start = s * seg_duration - seg_duration
        end = start + seg_duration

        if annotations is None:
            label = '[NULL]'
        else:
            label =  get_labels(prefix, start, end, annotations)

        out_name = "id_" + prefix + "_" + str(s) + "_l_" + label + ".wav"
        path_to_seg = os.path.join(save_to, out_name)    
        sig, rate = librosa.load(audio_file, sr=None, offset=start, duration=seg_duration)
        if verbose:
            print("Creating segment......", path_to_seg)
        librosa.output.write_wav(path_to_seg, sig, rate)

def _filter_annotations_by_filename(annotations, filename):
    """ Filter the annotations DataFrame by the base of the original file name (without the path or extension)

        Args:
        file: str
           The original audio file name without path or extensions.
           Ex: 'file_1' will match the entry './data/sample_a/file_1.wav" in the filename
           column of the annotations DataFrame.

        annotations: pandas.DataFrame
            DataFrame with the the annotations. At least the following columns are expected:
                "filename": the file name. Must be the the same as audio_file
                "label": the label value for each annotaded event
                "start": the start time relative to the beginning of the audio_file.
                "end": the end time relative to the beginning of the file.

        Returns:
            filtered annotations: pandas.DataFrame
            A subset of the annotations DataFrame containing only the entries for the specified file.
            
        Examples:
            >>> from ketos.data_handling.data_handling import _filter_annotations_by_filename
            >>> import pandas as pd
            >>> # Create an annotations dataframe
            >>> annotations = pd.DataFrame({'filename':['2min_01.wav','2min_01.wav','2min_02.wav','2min_02.wav','2min_02.wav'],
            ...                     'label':[1,2,1,1,1], 'start':[5.0, 100.5, 105.0, 80.0, 90.0],
            ...                     'end':[6.0,103.0,108.0, 87.0, 94.0]})
            >>> # Filter the annotations associated with file "2min_01"
            >>> annot_01 = _filter_annotations_by_filename(annotations,'2min_01')
            >>> annot_01
                  filename  label  start    end
            0  2min_01.wav      1    5.0    6.0
            1  2min_01.wav      2  100.5  103.0
                                 

    """
    filtered_indices = annotations.apply(axis=1, func= lambda row: os.path.basename(row.filename).split(".wav")[0] == filename)
    filtered_annotations = annotations[filtered_indices]
    return filtered_annotations

def get_labels(file, start, end, annotations, not_in_annotations=0):
    """ Retrieves the labels that fall in the specified interval.
    
        Args:
        file: str
           The base name (without paths or extensions) for the original audio file. Will be used to match the 'filename' field
           in the annotations Dataframe. Important: The name of the files must be
           unique within the annotations, even if the path is different.
           Ex: '/data/sample_a/file_1.wav' and '/data/sample_b/file_1.wav'

        annotations: pandas.DataFrame
            DataFrame with the the annotations. At least the following columns are expected:
                "filename": the file name. Must be the the same as audio_file
                "label": the label value for each annotaded event
                "start": the start time relative to the beginning of the audio_file.
                "end": the end time relative to the beginning of the file.
            
        not_in_annotations: str
            Label to be used if the segment is not included in the annotations.

        Returns:
            labels: str
                The labels corresponding to the interval specified.
                if the interval is not in the annotations, the value 
                specified in 'not_in_annotations' will be used.

        Examples:
            >>> from ketos.data_handling.data_handling import get_labels
            >>> import pandas as pd
            >>> audio_file="2min"
            >>> # Create an annotations dataframe
            >>> annotations = pd.DataFrame({'filename':['2min.wav','2min.wav','2min.wav'],
            ...                            'label':[1,2,1], 'start':[5.0, 100.5, 105.0],
            ...                            'end':[6.0,103.0,108.0]})
            >>> # Find all labels between time 4.0 seconds and time 5.0 seconds.
            >>> get_labels(file='2min',start=4.0, end=5.5,
            ...                        annotations=annotations, not_in_annotations=0)
            '[1]'
            >>>
            >>> # Find all labels between time 99.0 seconds and time 110.0 seconds.
            >>> get_labels(file='2min',start=99.0, end=110.0,
            ...                        annotations=annotations, not_in_annotations=0)
            '[2, 1]'

    """
    interval_start = start
    interval_end = end

    data = _filter_annotations_by_filename(annotations, file)
    query_results = data.query("(@interval_start >= start & @interval_start <= end) | (@interval_end >= start & @interval_end <= end) | (@interval_start <= start & @interval_end >= end)")
    #print(query_results)
    
    if query_results.empty:
        label = [not_in_annotations]
    else:
        label=[]
        for l in query_results.label:
          label.append(l)
                 
    return str(label)


def seg_from_time_tag(audio_file, start, end, name, save_to):
    """ Extract a segment from the audio_file according to the start and end tags.

        Args:
            audio_file:str
            .wav file name (including path).

            start:float
            Start point for segment (in seconds from start of the source audio_file)

            end:float
            End point for segment (in seconds from start of the source audio_file)
            
            save_to: str
            Path to the directory where segments will be saved.
            
            name: str
            Name of segment file name (including '.wav')
            

        Returns:

            None

        Examples:
            >>> import os
            >>> from ketos.data_handling.data_handling import read_wave
            >>> from ketos.data_handling.data_handling import seg_from_time_tag
            >>>
            >>> # Define the audio file and the destination folder
            >>> audio_file = "ketos/tests/assets/2min.wav"
            >>> save_dir = "ketos/tests/assets/tmp/segs_from_tags"
            >>> # Create the folder
            >>> os.makedirs(save_dir, exist_ok=True)
            >>>
            >>> # Create a segmet starting at 0.5 seconds and ending at 2.5 seconds
            >>> seg_from_time_tag(audio_file=audio_file, start=0.5, end=2.5 , name="seg_1.wav", save_to=save_dir )
            >>>
            >>> # Read the created segment and check its duration
            >>> rate, sig  = read_wave(os.path.join(save_dir, "seg_1.wav"))
            >>> duration = len(sig)/rate
            >>> duration
            2.0

    """
    out_seg = os.path.join(save_to, name)
    sig, rate = librosa.load(audio_file, sr=None, offset=start, duration=end - start)
    librosa.output.write_wav(out_seg, sig, rate)


def segs_from_annotations(annotations, save_to):
    """ Generates segments based on the annotations DataFrame.

        Args:
            annotations: pandas.DataFrame
            DataFrame with the the annotations. At least the following columns are expected:
                "filename": the file name. Must be the the same as audio_file
                "label": the label value for each annotaded event
                "start": the start time relative to the beginning of the audio_file.
                "end": the end time relative to the beginning of the file.
            save_to: str
            path to the directory where segments will be saved.
            
            
        Returns:
            None
            
        Examples:
            >>> import os
            >>> from glob import glob
            >>> import pandas as pd
            >>> from ketos.data_handling.data_handling import segs_from_annotations
            >>>
            >>> # Define the audio file and the destination folder
            >>> audio_file_path = "ketos/tests/assets/2min.wav"
            >>> save_dir = "ketos/tests/assets/tmp/from_annot"
            >>>
            >>> # Create a dataframe with annotations
            >>> annotations = pd.DataFrame({'filename':[audio_file_path,audio_file_path,audio_file_path],
            ...                            'label':[1,2,1], 'start':[5.0, 70.5, 105.0],
            ...                            'end':[6.0,73.0,108.0]})
            >>>
            >>> # Segemnt the audio file according with the annotations           
            >>> segs_from_annotations(annotations,save_dir)
            Creating segment...... ketos/tests/assets/tmp/from_annot id_2min_0_l_[1].wav
            Creating segment...... ketos/tests/assets/tmp/from_annot id_2min_1_l_[2].wav
            Creating segment...... ketos/tests/assets/tmp/from_annot id_2min_2_l_[1].wav
            
            

    """ 
    create_dir(save_to)
    for i, row in annotations.iterrows():
        start = row.start
        end= row.end
        base_name = os.path.basename(row.filename).split(".wav")[0]
        seg_name = "id_" + base_name + "_" + str(i) + "_l_[" + str(row.label) + "].wav"
        seg_from_time_tag(row.filename, row.start, row.end, seg_name, save_to)
        print("Creating segment......", save_to, seg_name)

def pad_signal(signal,rate, length):
    """Pad a signal with zeros so it has the specified length

        Zeros will be added before and after the signal in approximately equal quantities.
        Args:
            signal: numpy.array
            The signal to be padded

            rate: int
            The sample rate

            length: float
            The desired length for the signal

         Returns:
            padded_signal: numpy.array
            Array with the original signal padded with zeros.

        Examples:
            >>> from ketos.data_handling.data_handling import read_wave
            >>> from ketos.data_handling.data_handling import pad_signal
            >>>
            >>> #Read a very short audio signal
            >>> rate, sig = read_wave("ketos/tests/assets/super_short_1.wav")
            >>> # Calculate its duration (in seconds)
            >>> len(sig)/rate
            0.00075
            >>>
            >>> # Pad the signal
            >>> padded_signal = pad_signal(signal=sig, rate=rate, length=0.5)
            >>> # Now the duration is equal to the 0.5 seconds specified by the 'length' argument
            >>> len(padded_signal)/rate
            0.5
        
    """
    length = length * rate
    input_length = signal.shape[0] 
    
    difference = ( length - input_length) 
    pad1_len =  int(np.ceil(difference/2))
    pad2_len = int(difference - pad1_len)

    pad1 =  np.zeros((pad1_len))
    pad2 =  np.zeros((pad2_len))

    padded_signal =  np.concatenate([pad1,signal,pad2])
    return padded_signal


class AudioSequenceReader:
    """ Reads sequences of audio files and serves them as a merged :class::ketos.audio_processing.audio.AudioSignal.

        Separate audio files are joined together in a single time series. 

        If the source is given as a folder (rather than a list of file names), all wav files from
        the folder will be read.
        
        If the file names have date-time information, the date-time information will be 
        parsed and used to sort the files chronologically. Any gaps will be filled with zeros.

        If the file names do not contain date-time information, the files will be sorted 
        alphabetically (if given as a folder name) or in the order provided (if given as a list 
        of file names).

        OBS: Note that the AudioSequenceReader always loads entire wav files into memory, 
        even when the requested batch size is smaller than the file size. This can cause 
        problems if the wav files are very large.

        TODO: If the file size is larger than the batch size, only load the relevant part of 
        the wav file into memory, to avoid problems with very large wav files.
        
        Args:
            source: str or list
                File name, list of file names, or directory name 
            recursive_search: bool
                If True, includes .wav files from all subdirectories 
            rate: float
                Sampling rate in Hz
            datetime_stamp: str
                A default datetime to be used in case the file names do not contain datetime information.
                If left to None, the the AudioSequenceReader will try to extract datetime from the filename
                using the 'datetime_fmt' argument.
            datetime_fmt: str
                Format for parsing date-time data from file names. If the file names do not contain
                datetime information and a 'datetime_stamp' is not provided, the current day will be used
                a time starting at 00:00:00:000
            n_smooth: int
                Size of region (number of samples) used for smoothly joining audio signals 
            verbose: bool
                If True, print progress messages during processing
            batch_size_samples: int
                Number of samples loaded for each batch.
            batch_size_files: int
                Number of wav files loaded for each batch. (Overwrites batch_size_samples.)
        
        Raises:
            AssertionError:
                If the specified directory does not exist
                If the specified directory does not have any .wav files

        Examples:

        >>> from ketos.data_handling.data_handling import AudioSequenceReader
        >>> # Define the folder containing the audio files
        >>> path_to_files = "ketos/tests/assets/2s_segs"
        >>> # This folder contains 60 files (of 2 seconds each) 
        >>> # We want to read them in chunks of 10 (i.e.: 20-second long signals)       
        >>> 
        >>> # Define the size (in samples) for each batch.
        >>> size = 2000 * 20 # The sampling rate is 2000Hz
        >>> # Create an AudioSequenceReader object
        >>> reader = AudioSequenceReader(source=path_to_files, rate=2000)
        >>> # get the first audio signal
        >>> seq_1 = reader.next(size=size)
        >>> # The length the same as the specified size
        >>> len(seq_1.data)
        40000
        >>> seq_2 = reader.next(size=size)
        >>> len(seq_2.data)
        40000

    """
    def __init__(self, source, recursive_search=False, rate=None, datetime_stamp=None, datetime_fmt=None, n_smooth=100, verbose=False,\
            batch_size_samples=None, batch_size_files=None):
        self.rate = rate
        self.n_smooth = n_smooth
        self.times = list()
        self.files = list()
        self.batch = None
        self.signal = None
        self.index = -1
        self.time = None
        self.eof = False
        self.verbose = verbose
        self.batch_size_samples = batch_size_samples
        self.batch_size_files = batch_size_files
        self.load(source=source, recursive_search=recursive_search, datetime_stamp=datetime_stamp, datetime_fmt=datetime_fmt)

    def load(self, source, recursive_search=False, datetime_stamp=None, datetime_fmt=None):
        """
            Reset the reader and load new data.
            
            OBS: These method does not actually load any audio data into memory, but merely 
            creates a record of all the wav files to be processed.

            Args:
                source: str or list
                    File name, list of file names, or directory name 
                recursive_search: bool
                    If true, include wav files from all subdirectories
                datetime_stamp: str
                    A default datetime to be used in case the file names do not contain datetime information.
                    Requires the format to be specified by the 'datetime_fmt' argument.
                    If left to None, the the AudioSequenceReader will try to extract datetime from the filename
                    using the 'datetime_fmt'.
                datetime_fmt: str
                    Format for parsing date-time data from file names. If the file names do not contain
                    datetime information and a 'datetime_stamp' is not provided, the current day will be used
                    a time starting at 00:00:00:000

            Raises:
                AssertionError:
                    If the specified directory does not exist
                    If the specified directory does not have any .wav files

            Examples:

                >>> from glob import glob
                >>> from ketos.data_handling.data_handling import AudioSequenceReader
                >>>
                >>> # Define the folder containing the audio files
                >>> path_to_files = "ketos/tests/assets/2s_segs"
                >>> # Define a list with the 10 files
                >>> list_of_files_1 = glob(path_to_files + "/*" )[0:10]
                >>> # Define another list with 10 different files
                >>> list_of_files_2 = glob(path_to_files + "/*" )[10:20]
                >>>
                >>> # Define the size (in samples) for each batch.
                >>> size = 2000 * 20 # The sampling rate is 2000Hz
                >>>                
                >>> # Create an AudioSequenceReader object with the first list of files
                >>> reader = AudioSequenceReader(source=list_of_files_1, rate=2000)
                >>> # Load the reader with a new source of files
                >>> reader.load(source=list_of_files_1)

        """
        self.files.clear()

        # get list of file names
        fnames = list()
        if isinstance(source, list):
            fnames = source
        else:
            if source[-4:] == '.wav':
                fnames = [source]
            else:
                fnames = find_wave_files(path=source, subdirs=recursive_search)

            # sort file names alphabetically
            fnames = sorted(fnames)

        # check that files exist
        for f in fnames:
            assert os.path.exists(f), " Could not find {0}".format(f)

        # check that we have at least 1 file
        assert len(fnames) > 0, " No wave files found in {0}".format(source)

        # default time stamp
        if datetime_stamp is not None:
            t0 = datetime.datetime.strptime(datetime_stamp, datetime_fmt)  
        else:
            t0 = datetime.datetime.today()
            t0 = datetime.datetime.combine(t0, datetime.datetime.min.time())

        # time stamps
        for f in fnames:
            fmt = datetime_fmt
            p_unix = f.rfind('/')
            p_win = f.rfind('\\')
            p = max(p_unix, p_win)
            folder = f[:p+1]
            if folder is not None and fmt is not None:
                fmt = folder + fmt
            t = parse_datetime(f, fmt)
            if t is None:
                t = t0
            self.files.append([f, t])

        # sort signals in chronological order
        def sorting(y):
            return y[1]

        self.files.sort(key=sorting)

        # reset the reader
        self.reset()

    def _read_file(self, i):
        from ketos.audio_processing.audio import TimeStampedAudioSignal
    
        assert i < len(self.files), "attempt to read file with id {0} but only {1} files have been loaded".format(i, len(self.files))

        if self.verbose:
            print(' File {0} of {1}'.format(i+1, len(self.files)), end="\r")
            
        f = self.files[i]
        s = TimeStampedAudioSignal.from_wav(path=f[0], time_stamp=f[1]) # read in audio data from wav file
        
        if self.rate is not None:
            s.resample(new_rate=self.rate) # resamples

        return s
        
    def _read_next_file(self):
        """
            Read next file, increment file counter, and update time.
        """
        self.index += 1 # increment counter
        if self.index < len(self.files):
            self.signal = self._read_file(self.index) # read audio file
            self.time = self.signal.begin() # start time of this audio file
        else:
            self.signal = None
            self.time = None
            
        if self.signal is None:
            self.eof = True

    def _add_to_batch(self, size, new_batch):
        """
            Add audio from current file to batch.
            
            Args:
                size: int
                    Maximum batch size (number of samples) 
                new_batch: bool
                    Start a new batch or add to existing
        """
        if self.signal.empty():
            self._read_next_file()
           
        if self.signal is None:
            return

        file_is_new = self.signal.begin() == self.time # check if we have already read from this file

        if new_batch:
            if self.batch is not None:
                t_prev = self.batch.end() # end time of the previous batch
            else:
                t_prev = None

            self.batch = self.signal.split(s=size) # create a new batch

            if t_prev is not None and self.batch.begin() < t_prev: 
                self.batch.time_stamp = t_prev # ensure that new batch starts after the end of the previous batch
                
            if file_is_new: 
                self.times.append(self.batch.begin()) # collect times
        else:
            l = len(self.signal.data)
            if file_is_new:
                n_smooth = self.n_smooth
            else:
                n_smooth = 0

            t = self.batch.append(signal=self.signal, n_smooth=n_smooth, max_length=size) # add to existing batch
            if file_is_new and (self.signal.empty() or len(self.signal.data) < l): 
                self.times.append(t) # collect times
        
        if self.signal.empty() and self.index == len(self.files) - 1: # check if there is more data
            self.eof = True 

    def next(self, size=math.inf):
        """
            Read next batch of audio files and merge into a single audio signal. 
            
            If no maximum size is given, all loaded files will be read and merged, 
            unless either batch_size_samples or batch_size_files has been specified 
            at the time of initialization.  
            
            Args:
                size: int
                    Maximum batch size (number of samples).
                    
            Returns:
                batch: TimeStampedAudioSignal
                    Merged audio signal

            Examples:

                >>> from ketos.data_handling.data_handling import AudioSequenceReader
                >>>
                >>> # Define the folder containing the audio files
                >>> path_to_files = "ketos/tests/assets/2s_segs"
                >>> 
                >>> # Define the size (in samples) for each batch.
                >>> size = 2000 * 20 # The sampling rate is 2000Hz, so each batch will be 20s long
                >>> # Create an AudioSequenceReader object
                >>> reader = AudioSequenceReader(source = path_to_files, rate=2000)
                >>>
                >>> seq1 = reader.next(size=size)
                >>> len(seq1.data)
                40000
                >>> seq2 = reader.next(size=size)
                >>> len(seq2.data)
                40000
        """
        num_files = None

        # ensure that size has type int
        if size is not math.inf:
            size = int(size)
        
        elif self.batch_size_files is not None:
            num_files = self.batch_size_files

        elif self.batch_size_samples is not None:
            size = self.batch_size_samples

        if self.finished():
            return None
        
        # batch size corresponds to certain number of wav files
        if num_files is not None:
            file_no = 0
            while file_no < num_files and not self.finished():
                self._add_to_batch(size=math.inf, new_batch=(file_no==0))
                file_no += 1

        # batch size corresponds to certain number of samples
        else:
            length = 0        
            while length < size and not self.finished():
                self._add_to_batch(size=size, new_batch=(length==0))                
                if self.batch is not None:
                    length = len(self.batch.data)

        if self.finished() and self.verbose:
            print(' Successfully processed {0} files'.format(len(self.files)))

        return self.batch
        
                
    def finished(self):
        """
            Reader has read all load data.
            
            Returns: 
                x: bool
                True if all data has been process, False otherwise
            
            Example:

                >>> from ketos.data_handling.data_handling import AudioSequenceReader
                >>> # Define the folder containing the audio files
                >>> path_to_files = "ketos/tests/assets/2s_segs"
                >>> # This folder contains 60 files (of 2 seconds each) 
                >>> # We want to read them in chunks of 10 (i.e.: 20-second long signals)       
                >>> 
                >>> # Define the size (in samples) for each batch.
                >>> size = 2000 * 20 # The sampling rate is 2000Hz
                >>> # Create an AudioSequenceReader object
                >>> reader = AudioSequenceReader(source=path_to_files, rate=2000)
                >>>
                >>> # Go through the batches, retriving the merged ausio signal
                >>> # and printing the number of samples in each, until there are no btaches left unprocessed
                >>> while reader.finished() != True:
                ...     seq = reader.next(size=size)
                ...     print(len(seq.data))
                40000
                40000
                40000
                40000
                40000
                30300
                >>> # The last batch generated a signal with 30300 samples only, because there were not 
                >>> # enough files to create a 40000 samples signal as specified by the 'size' argument

            
        """
        return self.eof
    
    def reset(self):
        """
            Go back and start reading from the beginning of the first file.

            Examples:

                >>> from ketos.data_handling.data_handling import AudioSequenceReader
                >>> # Define the folder containing the audio files
                >>> path_to_files = "ketos/tests/assets/2s_segs"
                >>> 
                >>> # Define the size (in samples) for each batch.
                >>> size = 2000 * 60 # The sampling rate is 2000Hz, so each batch will be 60s long
                >>> # Create an AudioSequenceReader object
                >>> reader = AudioSequenceReader(source=path_to_files, rate=2000)
                
                >>> # Here we want 5 signals, even it it means they'll come from repeated batches.
                >>> 
                >>> for i in range(5):
                ...    seq = reader.next(size=size)
                ...    print(len(seq.data))
                ...    if reader.finished(): #When there are no batches left, reset the reader
                ...        reader.reset()
                120000
                110200
                120000
                110200
                120000


                
            
        """
        # reset 
        self.index = -1
        self.times.clear()
        self.eof = False
        self.batch = None

        # read the first file 
        self._read_next_file()

    def log(self):
        """
            Generate summary of all processed data.

            Returns:
                df: pandas DataFrame
                    Table with file names and time stamps

            Example:

                >>> from glob import glob
                >>> from ketos.data_handling.data_handling import AudioSequenceReader
                >>>
                >>> # Define the folder containing the audio files
                >>> path_to_files = "ketos/tests/assets/2s_segs"
                >>> # Define a list with the 10 files that start with 'id_2min_1'
                >>> list_of_files = glob(path_to_files + "/id_2min_1*" )
                >>> list_of_files.sort()
                >>> 
                >>> # Define the size (in samples) for each batch.
                >>> size = 2000 * 20 # The sampling rate is 2000Hz, so each batch will be 20s long
                >>> # Define a default datetime stamp
                >>> datetime_stamp = "1960-06-10 16:30:00.000"
                >>> # Define the datetime format specification
                >>> datetime_fmt = "%Y-%m-%d %H:%M:%S.%f"
                >>>
                >>> # Create an AudioSequenceReader object
                >>> reader = AudioSequenceReader(source = list_of_files, rate=2000, datetime_stamp=datetime_stamp, datetime_fmt=datetime_fmt)
                >>>
                >>> seq1 = reader.next(size=size)
                >>> seq2 = reader.next(size=size)
                >>> reader.log()
                                      time                                             file
                0  1960-06-10 16:30:00.000  ketos/tests/assets/2s_segs/id_2min_10_l_[0].wav
                1  1960-06-10 16:30:01.950  ketos/tests/assets/2s_segs/id_2min_11_l_[0].wav
                2  1960-06-10 16:30:03.900  ketos/tests/assets/2s_segs/id_2min_12_l_[0].wav
                3  1960-06-10 16:30:05.850  ketos/tests/assets/2s_segs/id_2min_13_l_[0].wav
                4  1960-06-10 16:30:07.800  ketos/tests/assets/2s_segs/id_2min_14_l_[0].wav
                5  1960-06-10 16:30:09.750  ketos/tests/assets/2s_segs/id_2min_15_l_[0].wav
                6  1960-06-10 16:30:11.700  ketos/tests/assets/2s_segs/id_2min_16_l_[0].wav
                7  1960-06-10 16:30:13.650  ketos/tests/assets/2s_segs/id_2min_17_l_[0].wav
                8  1960-06-10 16:30:15.600  ketos/tests/assets/2s_segs/id_2min_18_l_[0].wav
                9  1960-06-10 16:30:17.550  ketos/tests/assets/2s_segs/id_2min_19_l_[0].wav
                10 1960-06-10 16:30:19.500   ketos/tests/assets/2s_segs/id_2min_1_l_[0].wav

        """
        n = len(self.times)
        fnames = [x[0] for x in self.files]
        df = pd.DataFrame(data={'time':self.times,'file':fnames[:n]})
        return df


class AnnotationTableReader():
    """ Reads annotation data from csv file

        The csv file should have at least the following columns:

           * "filename": file name
           * "label": label value (0, 1, 2, ...)
           * "start": start time in seconds measured from the beginning of the audio file
           * "end": end time in seconds

        In addition, it can the following two optional columns:

           * "flow": frequency lower boundary in Hz
           * "fhigh": frequency upper boundary in Hz

        Args:
            path: str
                Full path to annotation csv file

        Example:

            >>> # create a pandas dataframe with dummy annotations
            >>> import pandas as pd
            >>> d = {'filename': ['x.wav', 'y.wav'], 'start': [3.0, 4.0], 'end': [17.0, 12.0], 'label': [0, 1]}
            >>> df = pd.DataFrame(data=d)
            >>> # save the dataframe to a csv file
            >>> fname = "ketos/tests/assets/tmp/ann.csv"
            >>> df.to_csv(fname)
            >>> # create an annotation reader
            >>> from ketos.data_handling.data_handling import AnnotationTableReader
            >>> reader = AnnotationTableReader(fname)
            >>> # get the annotations for x.wav
            >>> labels, boxes = reader.get_annotations('x.wav')
            >>> print(labels, boxes)
            [0] [[3.0, 17.0, 0.0, inf]]
    """
    def __init__(self, path):

        self.df = pd.read_csv(path)

        required_cols = ['filename', 'start', 'end', 'label']

        for r in required_cols:
            assert r in self.df.columns, 'column {0} is missing'.format(r)

        self.max_ann = None
        
    def get_max_annotations(self):
        """ Returns the maximum number of annotations attached to any single audio file.
            
            Returns:
                m: int
                    Number of annotations
        """        
        if self.max_ann is not None:
            m = self.max_ann
        else:
            x = self.df['filename'].values
            _, counts = np.unique(x, return_counts=True)
            m = np.max(counts)
            self.max_ann = m

        return m

    def get_annotations(self, filename):
        """ Get annotations associated with the specified file 
            
            Args:
                filename: str
                    File name
                    
            Returns:
                labels: list(int)
                    Labels
                boxes: list(tuple)
                    Boxes
        """
        # select rows with matching file name
        sel = self.df[self.df['filename']==filename]

        m = len(sel)

        # get labels
        labels = sel['label'].values

        # get times
        start = sel['start'].values
        stop  = sel['end'].values

        # get frequencies
        if 'flow' in sel.columns:
            flow = sel['flow'].values
        else:
            flow = np.zeros(m)
        if 'fhigh' in sel.columns:
            fhigh = sel['fhigh'].values        
        else:
            fhigh = np.ones(m) * math.inf

        # construct boxes
        boxes = list()
        for t1,t2,f1,f2 in zip(start, stop, flow, fhigh):
            b = [t1, t2, f1, f2]
            boxes.append(b)

        return labels, boxes


class SpecProvider():
    """ Compute spectrograms from raw audio (*.wav) files.

        Note that if spec_config is specified, the following arguments are ignored: 
        sampling_rate, window_size, step_size, length, overlap, flow, fhigh, cqt, bins_per_octave.

        TODO: Modify implementation so that arguments are not ignored when spec_config is specified.
    
        Args:
            path: str
                Full path to audio file (*.wav) or folder containing audio files
            channel: int
                For stereo recordings, this can be used to select which channel to read from
            spec_config: SpectrogramConfiguration
                Spectrogram configuration object.
            sampling_rate: float
                If specified, audio data will be resampled at this rate
            window_size: float
                Window size (seconds) used for computing the spectrogram
            step_size: float
                Step size (seconds) used for computing the spectrogram
            length: float
                Duration in seconds of individual spectrograms.
            overlap: float
                Overlap in seconds between consecutive spectrograms.
            flow: float
                Lower cut on frequency (Hz)
            fhigh: float
                Upper cut on frequency (Hz)
            cqt: bool
                Compute CQT magnitude spectrogram instead of the standard STFT magnitude 
                spectrogram.
            bins_per_octave: int
                Number of bins per octave. Only applicable if cqt is True.

            Example:
    """
    def __init__(self, path, channel=0, spec_config=None, sampling_rate=None, window_size=0.2, step_size=0.02, length=None,\
        overlap=0, flow=None, fhigh=None, cqt=False, bins_per_octave=32):

        if spec_config is None:
            spec_config = SpectrogramConfiguration(rate=sampling_rate, window_size=window_size, step_size=step_size,\
                bins_per_octave=bins_per_octave, window_function=None, low_frequency_cut=flow, high_frequency_cut=fhigh,\
                length=length, overlap=overlap, type=['Mag', 'CQT'][cqt])

        if spec_config.length is not None:
            assert spec_config.overlap < spec_config.length, 'Overlap must be less than spectrogram length'

        self.spec_config = spec_config
        self.channel = channel

        # get all wav files in the folder, including any subfolders
        if path[-3:].lower() == 'wav':
            self.files = [path]
        else:
            self.files = find_wave_files(path=path, fullpath=True, subdirs=True)

        # file ID
        self.fid = -1
        self._next_file()

    def __iter__(self):
        return self

    def __next__(self):
        """ Compute next spectrogram.

            Returns: 
                spec: instance of MagSpectrogram or CQTSpectrogram
                    Spectrogram.
        """
        # get spectrogram
        spec = self.get(time=self.time, file_id=self.fid)

        # increment time
        self.time += spec.duration() - self.spec_config.overlap

        # increment segment ID
        self.sid += 1

        # if this was the last segment, jump to the next file
        if self.sid == self.num_segs:
            self._next_file()

        return spec

    def reset(self):
        """ Go back to the beginning of the first file.
        """
        self.jump(0)

    def jump(self, file_id=0):
        """ Go to the beginning of the selected file.

            Args:
                file_id: int
                    File ID
        """
        self.fid = file_id - 1
        self._next_file()

    def get(self, time=0, file_id=0):
        """ Compute spectrogram from specific file and time.

            Args:
                time: float
                    Start time of the spectrogram in seconds, measured from the 
                    beginning of the file.
                file_id: int
                    Integer file identifier.
        
            Returns: 
                spec: instance of MagSpectrogram or CQTSpectrogram
                    Spectrogram.
        """
        from ketos.audio_processing.spectrogram import MagSpectrogram, CQTSpectrogram

        # file
        f = self.files[file_id]

        # compute spectrogram
        if self.spec_config.type == 'CQT':
            spec = CQTSpectrogram.from_wav(path=f, spec_config=self.spec_config,\
                offset=time, decibel=True, channel=self.channel)

        else:
            spec = MagSpectrogram.from_wav(path=f, spec_config=self.spec_config,\
                offset=time, decibel=True, adjust_duration=True, channel=self.channel)

        return spec

    def _next_file(self):
        """ Jump to next file. 
        """
        # increment file ID
        self.fid += 1

        if self.fid == len(self.files):
            self.fid = 0

        # check if file exists
        f = self.files[self.fid]
        exists = os.path.exists(f)

        # if not, jump to the next file
        if not exists:
            self._next_file()

        # get duration
        duration = librosa.get_duration(filename=f)

        if self.spec_config.length is None:
            self.num_segs = 1
        else:
            self.num_segs = int(np.ceil(duration / (self.spec_config.length - self.spec_config.overlap)))

        # reset segment ID and time
        self.sid = 0
        self.time = 0



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" CNN module within the ketos library

    This module provides utilities to work with Convolutional 
    Neural Networks for classification tasks.

    Contents:
        DataHandler class:
        DataUse class:
"""

import tensorflow as tf
import numpy as np
import pandas as pd
from collections import namedtuple
from ketos.neural_networks.neural_networks import DataHandler, DataUse, predictions, class_confidences
from ketos.data_handling.data_handling import from1hot, to1hot, get_image_size
from ketos.data_handling.data_feeding import ActiveLearningBatchGenerator


ConvParams = namedtuple('ConvParams', 'name n_filters filter_shape')
ConvParams.__doc__ = '''\
Name and dimensions of convolutional layer in neural network

name - Name of convolutional layer, e.g. "conv_layer"
n_filters - Number of filters, e.g. 16
filter_shape - Filter shape, e.g. [4,4]'''


class BasicCNN(DataHandler):
    """ Convolutional Neural Network for classification tasks.

        The network architecture can only be constructed if the shape of the input data is known.
        Therefore, either image_shape or train_x must be provided.
        In the latter case, the shape is automatically determined from the input data.

        Args:
            image_shape: tuple
                Shape of input images
            train_x: numpy array
                Training data
            train_y: numpy array
                Labels for training data. Can be integers or 1-hot encoded
            validation_x: numpy array
                Validation data
            validation_y: numpy array
                Labels for validation data. Can be integers or 1-hot encoded
            test_x: numpy array
                Test data
            test_y: numpy array
                Labels for test data. Can be integers or 1-hot encoded
            num_labels: int
                Number of labels
            batch_size: int
                The number of examples in each batch
            num_epochs: int
                The number of epochs
            learning_rate: float
                The learning rate to be using by the optimization algorithm
            keep_prob: float
                Probability of keeping weights during training. Set keep_prob to 1.0 to disable drop-out (default).
            seed: int
                Seed to be used by both tensorflow and numpy when generating random numbers            
            verbosity: int
                Verbosity level (0: no messages, 1: warnings only, 2: warnings and diagnostics)
            max_frames: int
                Maximum number of frames that will be acted on by a tensorflow operation at a time.
                Useful to avoid memory warnings/errors.

        Attributes:
            learning_rate_value: float
                The learning rate to be using by the optimization algorithm
            keep_prob_value: float
                Probability of keeping weights during training. Set keep_prob to 1.0 to disable drop-out (default).
            sess: tf.Session
                Current TensorFlow session
            epoch_counter: int
                Keeps count of the current training epoch
            image_shape: tuple
                Shape of input data                

        Example:

            >>> # initialize BasicCNN for classifying 2x2 images
            >>> from ketos.neural_networks.cnn import BasicCNN
            >>> cnn = BasicCNN(image_shape=(2,2))
            >>> print(cnn.image_shape)
            (2, 2)
    """
    def __init__(self, image_shape=None, train_x=None, train_y=None, 
                validation_x=None, validation_y=None, test_x=None, test_y=None, 
                num_labels=2, batch_size=128, num_epochs=10, learning_rate=0.01, 
                keep_prob=1.0, seed=42, verbosity=2, max_frames=1E4):

        assert (image_shape is not None) or (train_x is not None), "image_shape or train_x must be provided"

        self.batch_size = batch_size
        self.num_epochs = num_epochs
        self.learning_rate_value = learning_rate
        self.keep_prob_value = keep_prob
        self.set_seed(seed)
        self.verbosity = verbosity
        self.max_frames = max_frames

        self.sess = tf.Session()
        self.epoch_counter = 0

        super(BasicCNN, self).__init__(train_x=train_x, train_y=train_y, 
                validation_x=validation_x, validation_y=validation_y,
                test_x=test_x, test_y=test_y, num_labels=num_labels)

        if train_x is not None:
            self.image_shape = self._image_shape()
        else:
            self.image_shape = image_shape

    def reset(self):
        """ Reset the state of the network.

            Resets the epoch counter to 0 and reinitializes all weights to random values.
        """
        self.epoch_counter = 0
        self.sess.run(self.init_op)

    def set_tf_nodes(self, tf_nodes, reset=True):
        """ Set the nodes of the tensorflow graph as instance attributes, so that other methods can access them

            Args:
                tf_nodes: dict
                    A dictionary with the tensorflow objects necessary to train and run the model:
                    sess, x, y, cost_function, optimizer, predict, correct_prediction,
                    accuracy, init_op, merged, writer, saver, keep_prob, learning_rate, class_weights.
                    These objects are stored as instance attributes when the class is instantiated.
                reset: bool
                    Reset the epoch counter and re-initialize the weights

            Returns:
                None
        """
        self.x = tf_nodes['x']
        self.y = tf_nodes['y']
        self.cost_function = tf_nodes['cost_function']
        self.optimizer = tf_nodes['optimizer']
        self.predict = tf_nodes['predict']
        self.correct_prediction = tf_nodes['correct_prediction']
        self.accuracy = tf_nodes['accuracy']
        self.init_op = tf_nodes['init_op']
        self.merged = tf_nodes['merged']
        self.writer = tf_nodes['writer']
        self.saver = tf_nodes['saver']
        self.keep_prob = tf_nodes['keep_prob']
        self.learning_rate = tf_nodes['learning_rate']
        self.class_weights = tf_nodes['class_weights']
        self.is_train = tf_nodes['is_train']

        if reset:
            self.reset()

    def set_seed(self,seed):
        """Set the random seed.

            Useful to generate reproducible results.

            Args:
                seed: int
                    Seed to be used by both tensorflow and numpy when generating random numbers
            Returns:
                None        
        """
        np.random.seed(seed)
        tf.set_random_seed(seed)
             
    def set_verbosity(self, verbosity):
        """Set verbosity level.
            
            0: no messages
            1: warnings only
            2: warnings and diagnostics

            Args:
                verbosity: int
                    Verbosity level.        
        """
        self.verbosity = verbosity

    def load(self, saved_meta, checkpoint, reset=False):
        """Load the Neural Network structure and weights from a saved model.

            See the save() method. 

            Args:
                saved_meta: str
                    Path to the saved .meta file.

                checkpoint: str
                    Path to the checkpoint to be used when loading the saved model

                reset: bool
                    Reset weights or use weights from the saved model.

            Returns:
                tf_nodes: dict
                    A dictionary with the tensorflow objects necessary to train and run the model:
                    sess, x, y, cost_function, optimizer, predict, correct_prediction,
                    accuracy, init_op, merged, writer, saver, keep_prob, learning_rate, class_weights.
        """
        sess = self.sess
        restorer = tf.train.import_meta_graph(saved_meta)
        restorer.restore(sess, tf.train.latest_checkpoint(checkpoint))

        graph = tf.get_default_graph()
        x = graph.get_tensor_by_name("x:0")
        y = graph.get_tensor_by_name("y:0")
        cost_function = graph.get_tensor_by_name("cost_function:0")
        optimizer = graph.get_operation_by_name("optimizer")
        predict = graph.get_tensor_by_name("predict:0")
        correct_prediction = graph.get_tensor_by_name("correct_prediction:0")
        accuracy = graph.get_tensor_by_name("accuracy:0")
        keep_prob = graph.get_tensor_by_name("keep_prob:0")        
        learning_rate = graph.get_tensor_by_name("learning_rate:0")      
        class_weights = graph.get_tensor_by_name("class_weights:0")  
        is_train = graph.get_tensor_by_name("is_train:0")  

        init_op = tf.global_variables_initializer()
        merged = tf.summary.merge_all()
        writer = tf.summary.FileWriter('summaries')
        saver = tf.train.Saver()

        tf_nodes = {'x': x,
                'y':y,            
                'cost_function': cost_function,
                'optimizer': optimizer,
                'predict': predict,
                'correct_prediction':correct_prediction,
                'accuracy': accuracy,
                'init_op': init_op,
                'merged': merged,
                'writer': writer,
                'saver': saver,
                'keep_prob': keep_prob,
                'learning_rate': learning_rate,
                'class_weights': class_weights,
                'is_train': is_train,
                }

        self.set_tf_nodes(tf_nodes=tf_nodes, reset=reset)

        return tf_nodes

    def create(self, conv_params=[ConvParams(name='conv_1',n_filters=32,filter_shape=[2,8]),\
            ConvParams(name='conv_2',n_filters=64,filter_shape=[30,8])], dense_size=[512],\
            batch_norm = False):
        """Create the Neural Network structure.

            The Network has a number of convolutional layers followed by a number 
            of fully connected layers with ReLU activation functions and a final 
            output layer with softmax activation.

            Each new convolutional layer is created with the method _create_new_conv_layer().

            Note that the window used for performing max-pooling on each convolutional layer has 
            the shape [2,2].
            
            The default network structure has two convolutional layers 
            and one fully connected layers with ReLU activation.

            Args:
                conv_params: list(ConvParams)
                    Configuration parameters for the convolutional layers.
                    Each item in the list represents a convolutional layer.
                dense_size: list(int)
                    Sizes of the fully connected layers preceeding the output layer.
                batch_norm: bool
                    Add batch normalization layers

            Returns:
                tf_nodes: dict
                    A dictionary with the tensorflow objects necessary to train and run the model:
                    sess, x, y, cost_function, optimizer, predict, correct_prediction,
                    accuracy, init_op, merged, writer, saver, keep_prob, learning_rate, class_weights.

            Example:

                >>> # initialize BasicCNN for classifying 4x4 images
                >>> from ketos.neural_networks.cnn import BasicCNN
                >>> cnn = BasicCNN(image_shape=(4,4), verbosity=2)
                >>> # create a small network with two convolutional layers and one dense layer
                >>> layer1 = ConvParams(name='conv_1', n_filters=4, filter_shape=[2,2])
                >>> layer2 = ConvParams(name='conv_2', n_filters=8, filter_shape=[4,4])
                >>> _ = cnn.create(conv_params=[layer1, layer2], dense_size=[8])
                ======================================================
                                   Convolutional layers               
                ------------------------------------------------------
                  Name   Input x Filters   Filter Shape   Output dim. 
                ------------------------------------------------------
                  conv_1       1 x 4          [2,2]         16
                  conv_2       4 x 8          [4,4]         8
                ======================================================
                                  Fully connected layers              
                ------------------------------------------------------
                  Name       Size                                      
                ------------------------------------------------------
                  dense_1    8
                  class_weights    2
                ======================================================
        """
        input_shape = self._image_shape()
        num_labels = self.num_labels
        
        if len(input_shape) <= 2:
            num_channels = 1
        else:
            num_channels = input_shape[2]

        keep_prob = tf.placeholder(tf.float32, name='keep_prob')
        learning_rate = tf.placeholder(tf.float32, name='learning_rate')

        is_train = tf.placeholder_with_default(False, (), 'is_train')

        if num_channels == 1:
            x = tf.placeholder(tf.float32, [None, input_shape[0] * input_shape[1]], name="x")
        else:
            x = tf.placeholder(tf.float32, [None, input_shape[0] * input_shape[1], num_channels], name="x")

        x_shaped = tf.reshape(x, [-1, input_shape[0], input_shape[1], num_channels])

        y = tf.placeholder(tf.float32, [None, num_labels], name="y")

        pool_shape = [2,2]

        # input and convolutional layers
        params = [ConvParams(name='input', n_filters=num_channels, filter_shape=[1,1])] # input layer with dimension 1
        params.extend(conv_params)
            
        # dense layers including output layer
        dense_size.append(num_labels)

        # batch normalisation on input
        if batch_norm:
            x_shaped = tf.layers.batch_normalization(x_shaped, training=is_train, momentum=0.99)
            update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
            with tf.control_dependencies(update_ops):
                x_shaped = tf.identity(x_shaped) # this ensures that moving average and variance are computed automatically (in training mode)
    
        # create convolutional layers
        conv_layers = [x_shaped]
        N = len(params)
        conv_summary = list()
        for i in range(1,N):
            # previous layer
            l_prev = conv_layers[len(conv_layers)-1]
            # layer parameters
            n_input = params[i-1].n_filters
            n_filters = params[i].n_filters
            filter_shape = params[i].filter_shape
            name = params[i].name
            # create new layer
            l = self._create_new_conv_layer(l_prev, n_input, n_filters, filter_shape, pool_shape, name, batch_norm, is_train)
            conv_layers.append(l)
            # collect info
            dim = l.shape[1] * l.shape[2] * l.shape[3]
            conv_summary.append("  {0}       {1} x {2}          [{3},{4}]         {5}".format(name, n_input, n_filters, filter_shape[0], filter_shape[1], dim))
            # apply drop-out 
            drop_out = tf.nn.dropout(l, keep_prob)
            conv_layers.append(drop_out)

        # last layer
        last = conv_layers[-1]

        # flatten
        dim = last.shape[1] * last.shape[2] * last.shape[3]
        flattened = tf.reshape(last, [-1, dim])

        # fully-connected layers with ReLu activation
        dense_layers = [flattened]
        dense_summary = list()
        for i in range(len(dense_size)):
            size = dense_size[i] 
            l_prev = dense_layers[i]
            w_name = 'w_{0}'.format(i+1)
            w = tf.Variable(tf.truncated_normal([int(l_prev.shape[1]), size], stddev=0.03), name=w_name)
            b_name = 'b_{0}'.format(i+1)
            b = tf.Variable(tf.truncated_normal([size], stddev=0.01), name=b_name)
            l = tf.matmul(l_prev, w) + b
            if i < len(dense_size) - 1:
                n = 'dense_{0}'.format(i+1)
                l = tf.nn.relu(l, name=n) # ReLu activation
            else: # output layer
                losses = tf.nn.softmax_cross_entropy_with_logits(logits=l, labels=y)
                cross_entropy = tf.reduce_mean(losses, name="cost_function")
                n = 'class_weights'
                l = tf.nn.softmax(l, name=n) # softmax                    

            dense_layers.append(l)
            dense_summary.append("  {0}    {1}".format(n, size))

        if self.verbosity >= 2:
            print('======================================================')
            print('                   Convolutional layers               ')
            print('------------------------------------------------------')
            print('  Name   Input x Filters   Filter Shape   Output dim. ')
            print('------------------------------------------------------')
            for line in conv_summary:
                print(line)
            print('======================================================')
            print('                  Fully connected layers              ')
            print('------------------------------------------------------')
            print('  Name       Size                                      ')
            print('------------------------------------------------------')
            for line in dense_summary:
                print(line)
            print('======================================================')

        # output layer
        y_ = dense_layers[-1]

        # add an optimizer
        optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate, name="optimizer").minimize(cross_entropy)

        # define an accuracy assessment operation
        predict = tf.argmax(y_, 1, name="predict")
        correct_prediction = tf.equal(tf.argmax(y, 1), tf.argmax(y_, 1), name="correct_prediction")
        accuracy = tf.reduce_mean(tf.cast(correct_prediction, tf.float32), name="accuracy")

        # setup the initialisation operator
        init_op = tf.global_variables_initializer()

        # setup recording variables
        # add a summary to store the accuracy
        tf.summary.scalar('accuracy', accuracy)
        merged = tf.summary.merge_all()

        writer = tf.summary.FileWriter('summaries')
        saver = tf.train.Saver()

        tf_nodes = {'x': x,
                'y':y,            
                'cost_function': cross_entropy,
                'optimizer': optimizer,
                'predict': predict,
                'correct_prediction': correct_prediction,
                'accuracy': accuracy,
                'init_op': init_op,
                'merged': merged,
                'writer': writer,
                'saver': saver,
                'keep_prob': keep_prob,
                'learning_rate': learning_rate,
                'class_weights': y_,
                'is_train': is_train
                }

        self.set_tf_nodes(tf_nodes)

        return tf_nodes
        
    def _create_new_conv_layer(self, input_data, num_input_channels, num_filters, filter_shape, pool_shape, name, batch_norm, is_train):
        """Create a convolutional layer.

            Args:
                input_data: tensorflow tensor
                    The input nodes for the convolutional layer
                num_input_channels: int
                    The number of input channels in input image
                num_filters: int
                    Number of filters to be used in the convolution
                filter_shape: list (int)
                    List of integers defining the shape of the filters.
                    Example: [2,8]
                pool_shape: list (int)
                    List of integers defining the shape of the pooling window.
                    Example: [2,8]
                name: str
                    Name by which the layer will be identified in the graph.
                batch_norm: bool
                    Add batch normalization layer
                is_train: tensorflow bool tensor
                    Boolean variables indicating if network is being trained or used for prediction

            Returns:
                out_layer: tensorflow layer
                    The convolutional layer.

        """
        # setup the filter input shape for tf.nn.conv_2d
        conv_filt_shape = [filter_shape[0], filter_shape[1], num_input_channels, num_filters]

        # initialise weights and bias for the filter
        weights = tf.Variable(tf.truncated_normal(conv_filt_shape, stddev=0.03), name=name+'_W')
        bias = tf.Variable(tf.truncated_normal([num_filters]), name=name+'_b')

        # setup the convolutional layer operation
        out_layer = tf.nn.conv2d(input_data, weights, [1, 1, 1, 1], padding='SAME')

        # add the bias
        out_layer += bias

        # apply a ReLU non-linear activation
        out_layer = tf.nn.relu(out_layer)

        # apply batch normalization 
        if batch_norm:
            out_layer = tf.layers.batch_normalization(out_layer, training=is_train, momentum=0.99)
            update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
            with tf.control_dependencies(update_ops):
                out_layer = tf.identity(out_layer)

        # now perform max pooling
        # ksize is the argument which defines the size of the max pooling window (i.e. the area over which the maximum is
        # calculated).  It must be 4D to match the convolution - in this case, for each image we want to use a 2 x 2 area
        # applied to each channel
        ksize = [1, pool_shape[0], pool_shape[1], 1]
        # strides defines how the max pooling area moves through the image - a stride of 2 in the x direction will lead to
        # max pooling areas starting at x=0, x=2, x=4 etc. through your image.  If the stride is 1, we will get max pooling
        # overlapping previous max pooling areas (and no reduction in the number of parameters).  In this case, we want
        # to do strides of 2 in the x and y directions.
        strides = [1, 2, 2, 1]
        out_layer = tf.nn.max_pool(out_layer, ksize=ksize, strides=strides, padding='SAME')

        return out_layer

    def _image_shape(self):
        """ Get the shape of the input data.

            Returns:
                img_shape: tuple
                    Shape of the input data images

        """
        tr = self.images[DataUse.TRAINING]

        if tr is None:
            img_shape = self.image_shape

        else:        

            img_shape = get_image_size(tr) 
            self.image_shape = img_shape

            if self.images[DataUse.VALIDATION] is not None:
                assert img_shape == get_image_size(self.images[DataUse.VALIDATION]), "Training and validation images must have same shape."

            if self.images[DataUse.TEST] is not None:
                assert img_shape == get_image_size(self.images[DataUse.TEST]), "Training and test images must have same shape."

        return img_shape

    def train(self, train_batch_gen=None,  batch_size=None, num_epochs=None, learning_rate=None, keep_prob=None, val_acc_goal=1.01, log_path=None):
        """ Train the neural network on the training set.

            Train on the batches of training data provided by the generator ``train_batch_gen``. 

            If a generator is not provided, train on the data stored in the ``images``
            attribute, dividing these data into batches of size ``batch_size``.

            Once training is done, check the accuracy on the validation set.
            
            Record summary statics during training. 


            Args:
                train_batch_gen:
                    A generator that provides (X,Y) tuples, whereeach tuple is
                    a batch of training data (X) and the corresponding labels (Y).
                batch_size: int
                    Batch size. Overwrites batch size specified at initialization.
                num_epochs: int
                    Number of epochs: Overwrites number of epochs specified at initialization.
                learning_rate: float
                    The learning rate to be using by the optimization algorithm
                keep_prob: float
                    Float in the range [0,1] specifying the probability of keeping the weights, 
                    i.e., drop-out will be applied if keep_prob < 1.
                val_acc_goal: float 
                    Terminate training when validation accuracy reaches this value 
                log_path: str
                    Path to file where training and validation accuracies will be saved for each epoch

            Returns:
                train_acc_epoch: list
                    Accuracy on training data for each epoch
                val_acc_epoch: list
                    Accuracy on validation data for each epoch. Will be 0 if no validation data is provided.

            Example:

                >>> # initialize BasicCNN for classifying 2x2 images
                >>> from ketos.neural_networks.cnn import BasicCNN
                >>> cnn = BasicCNN(image_shape=(2,2), verbosity=0, seed=1)
                >>> # create a small network with one convolutional layers and one dense layer
                >>> params = ConvParams(name='conv_1', n_filters=4, filter_shape=[2,2])
                >>> _ = cnn.create(conv_params=[params], dense_size=[4])
                >>> # give the network some training data
                >>> img0 = np.zeros(shape=(2,2))
                >>> img1 = np.ones(shape=(2,2))
                >>> x = [img0, img1, img0, img1, img0] # input data
                >>> y = [0, 1, 0, 1, 0] # labels
                >>> cnn.set_training_data(x, y)
                >>> cost, _ = cnn.train(batch_size=2, num_epochs=7, learning_rate=0.005)
        """
        if batch_size is None:
            batch_size = self.batch_size
        if num_epochs is None:
            num_epochs = self.num_epochs
        if keep_prob is None:
            keep_prob = self.keep_prob_value
        if learning_rate is None:
            learning_rate = self.learning_rate_value

        sess = self.sess

        x, y = self.get_training_data()
        x_val, y_val = self.get_validation_data()

        y = self._ensure1hot(y)
        y_val = self._ensure1hot(y_val)

        self.writer.add_graph(sess.graph)

        if self.verbosity >= 2:
            if self.epoch_counter == 0: 
                print("\nTraining  started")
            header = '\nEpoch  Cost  Train acc.  Val acc.'
            line   = '----------------------------------'
            print(header)
            print(line)

        # initialise the variables
        if self.epoch_counter == 0:
            sess.run(self.init_op)

        if train_batch_gen is not None:
            batches = train_batch_gen.n_batches
        else:
            batches = int(y.shape[0] / batch_size)

        train_acc_epoch = list()
        val_acc_epoch = list()

        for epoch in range(num_epochs):
            avg_cost = 0
            avg_acc = 0
            val_acc = 0
            for i in range(batches):
                if train_batch_gen is None:
                    offset = i * batch_size
                    x_i = x[offset:(offset + batch_size), :, :, :]
                   
                    y_i = y[offset:(offset + batch_size)]
                else:
                    x_i, y_i = next(train_batch_gen)

                x_i = self._reshape_x(x_i)

                fetch = [self.optimizer, self.cost_function, self.accuracy]

                _, c, a = sess.run(fetches=fetch, feed_dict={self.x: x_i, self.y: y_i, self.learning_rate: learning_rate, self.keep_prob: keep_prob, self.is_train: True})
                
                avg_cost += c / batches
                avg_acc += a / batches

            val_acc = self.accuracy_on_validation()

            if self.verbosity >= 2:
                s = ' {0}/{4}  {1:.3f}  {2:.3f}  {3:.3f}'.format(epoch + 1, avg_cost, avg_acc, val_acc, num_epochs)
                print(s)

            summary = self._get_summary(x=x_val, y=y_val)
            if summary is not None:
                self.writer.add_summary(summary, self.epoch_counter)

            self.epoch_counter += 1

            if log_path is not None:
                if self.epoch_counter == 1:
                    log_file = open(log_path, 'w+')
                else:
                    log_file = open(log_path, 'a')

                log_file.write('{0},{1:.4f},{2:.4f}\n'.format(self.epoch_counter, avg_acc, val_acc))
                log_file.close()

            train_acc_epoch.append(avg_acc)
            val_acc_epoch.append(val_acc)

            if val_acc >= val_acc_goal:
                break

        if self.verbosity >= 2:
            print(line)

        return train_acc_epoch, val_acc_epoch

    def train_active(self, provider, num_sessions=1, num_epochs=None, learning_rate=None, keep_prob=None, val_acc_goal=1.01, log_path=None):
        """ Train the neural network in an active manner using a data provider module.

            Args:
                provider: ActiveLearningBatchGenerator
                    ActiveLearningBatchGenerator
                num_sessions: int
                    Number of training iterations.
                batch_size: int
                    Batch size. Overwrites batch size specified at initialization.
                num_epochs: int
                    Number of epochs: Overwrites number of epochs specified at initialization.
                learning_rate: float
                    The learning rate to be using by the optimization algorithm
                keep_prob: float
                    Float in the range [0,1] specifying the probability of keeping the weights, 
                    i.e., drop-out will be applied if keep_prob < 1.
                val_acc_goal: float 
                    Terminate training when validation accuracy reaches this value 
                log_file: str
                    Path to file where training and validation accuracies will be saved for each epoch

            Returns:
                train_acc_epoch: list
                    Accuracy on training data for each epoch
                val_acc_epoch: list
                    Accuracy on validation data for each epoch. Will be 0 if no validation data is provided.

            Example:

                >>> # initialize BasicCNN for classifying 2x2 images
                >>> from ketos.neural_networks.cnn import BasicCNN
                >>> cnn = BasicCNN(image_shape=(2,2), verbosity=0, seed=1)
                >>> # create a small network with one convolutional layers and one dense layer
                >>> params = ConvParams(name='conv_1', n_filters=4, filter_shape=[2,2])
                >>> _ = cnn.create(conv_params=[params], dense_size=[4])
                >>> # create some training data
                >>> img0 = np.zeros(shape=(2,2))
                >>> img1 = np.ones(shape=(2,2))
                >>> x = [img0, img1, img0, img1, img0, img1, img0] # input data
                >>> y = [0, 1, 0, 1, 0, 1, 0] # labels
                >>> # create a data provider
                >>> from ketos.data_handling.data_feeding import ActiveLearningBatchGenerator
                >>> g = ActiveLearningBatchGenerator(session_size=4, batch_size=2, x=x, y=y)
                >>> cost, _ = cnn.train_active(provider=g, num_sessions=3, num_epochs=7, learning_rate=0.005)
        """    
        provider.return_indices = False # ensure that the batch generator only returns x,y
        provider.convert_to_one_hot = True 

        # set validation data, if any
        X, Y = provider.get_validation_data()
        if len(X) > 0:
            self.set_validation_data(x=X, y=Y)

        train_acc_epoch = list()
        val_acc_epoch = list()

        for i in range(num_sessions):

            if self.verbosity >= 2:
                print('\nSession: {0}/{1}'.format(i+1, num_sessions))

            # batch generator
            gen = next(provider)

            # train
            train_acc, val_acc = self.train(train_batch_gen=gen, num_epochs=num_epochs,\
                    learning_rate=learning_rate, keep_prob=keep_prob, val_acc_goal=val_acc_goal,\
                    log_path=log_path)

            # update predictions and confidences
            num_batches = gen.n_batches
            gen.return_batch_ids = True
            for _ in range(num_batches):
                i, x, _ = next(gen)
                w = self.get_class_weights(x)
                pred = predictions(w)
                conf = class_confidences(w)
                provider.update_performance(indices=i, predictions=pred, confidences=conf)

            if val_acc[-1] >= val_acc_goal:
                break

            train_acc_epoch += train_acc
            val_acc_epoch += val_acc

        return train_acc_epoch, val_acc_epoch

    def save(self, destination):
        """ Save the model

            Args:
                destination: str
                    Path to the file in which the model will be saved. 

            Returns:
                None.
        """
        self.saver.save(self.sess, destination)

    def _get_summary(self, x, y):
        """ Obtain summary of model performance on dataset x, y

            Obs: If the dataset contains more than max_frames samples, only the first 
            max_frames samples will be computed.

            Args:
                x: tensor
                    Tensor containing the input data
                y: tensor
                    Tensor containing the labels

            Returns:
                summary
        """

        if x is None:
            return None

        x = self._reshape_x(x)

        N = min(x.shape[0], self.max_frames)
        x = x[:N]
        y = y[:N]

        summary = self.sess.run(fetches=self.merged, feed_dict={self.x: x, self.y: y, self.learning_rate: self.learning_rate_value, self.keep_prob: 1.0})

        return summary

    def _check_accuracy(self, x, y):
        """ Check accuracy of the model by checking how close
            to y the models predictions are when fed x

            Based on the accuracy operation stored in the attribute 'self.accuracy'),
            which is defined by the 'create_net_structure()' method.

            Args:
                x: tensor
                    Tensor containing the input data
                y: tensor
                    Tensor containing the labels

            Returns:
                results: float
                    The accuracy value
        """
        if x is None:
            return 0

        x = self._reshape_x(x) 
        y1hot = self._ensure1hot(y)       

        x, y1hot = self._split(x, y1hot)
        results = list()
        for xi,yi in zip(x, y1hot):
            r = self.sess.run(fetches=self.accuracy, feed_dict={self.x:xi, self.y:yi, self.learning_rate: self.learning_rate_value, self.keep_prob:1.0})
            results.append(r)

        results = np.array(results)
        results = np.average(results)

        return results

    def get_predictions(self, x):
        """ Predict labels by running the model on x

            Args:
                x: tensor
                    Tensor containing the input data.
                
            Returns:
                results: vector
                    A vector containing the predicted labels.                

            Example:

                >>> # initialize BasicCNN for binary classification of 2x2 images
                >>> from ketos.neural_networks.cnn import BasicCNN
                >>> cnn = BasicCNN(image_shape=(2,2), verbosity=0, seed=1, num_labels=2)
                >>> # create a small network with one convolutional layers and one dense layer
                >>> params = ConvParams(name='conv_1', n_filters=4, filter_shape=[2,2])
                >>> _ = cnn.create(conv_params=[params], dense_size=[4])
                >>> # create a 2x2 image
                >>> img = np.zeros(shape=(2,2))
                >>> # obtain the label predicted by the untrained network 
                >>> p = cnn.get_predictions(img)
        """
        x = self._reshape_x(x)

        x, _ = self._split(x)
        results = list()
        for xi in x:
            r = self.sess.run(fetches=self.predict, feed_dict={self.x:xi, self.learning_rate: self.learning_rate_value, self.keep_prob:1.0})
            results = results + r.tolist()

        results = np.array(results)
        return results

    def get_features(self, x, layer_name):
        """ Compute feature vector by running the model on x

            Args:
                x: tensor
                    Tensor containing the input data.
                layer_name: str
                    Name of the feature layer.
                
            Returns:
                results: vector
                    A vector containing the feature values.                
        """
        x = self._reshape_x(x)

        graph = tf.get_default_graph()
        f = graph.get_tensor_by_name("{0}:0".format(layer_name)) 

        x, _ = self._split(x)
        results = list()
        for xi in x:
            r = self.sess.run(fetches=f, feed_dict={self.x:xi, self.learning_rate: self.learning_rate_value, self.keep_prob:1.0})
            results = results + r.tolist()

        results = np.array(results)
        return results

    def get_class_weights(self, x):
        """ Compute classification weights by running the model on x.

            Args:
                x:tensor
                    Tensor containing the input data.
                
            Returns:
                results: vector
                    A vector containing the classification weights. 
        """
        x = self._reshape_x(x)

        fetch = self.class_weights

        x, _ = self._split(x)
        results = list()
        for xi in x:
            feed = {self.x:xi, self.learning_rate: self.learning_rate_value, self.keep_prob:1.0}
            r = self.sess.run(fetches=fetch, feed_dict=feed)
            results = results + r.tolist()

        results = np.array(results)
        return results

    def _reshape_x(self, x):
        """ Reshape input data from a 2d matrix to a 1d vector.

            Args:
                x: numpy array
                    2d array containing the input data.
            
            Returns:
                results: vector
                    A vector containing the flattened inputs.                
        """
        img_shape = self._image_shape()

        if len(img_shape) == 1:
            reshaped_x = np.reshape(x, (-1, img_shape[0]))

        elif (len(img_shape) == 2) or (len(img_shape) > 2 and img_shape[2] == 1):
            reshaped_x = np.reshape(x, (-1, img_shape[0] * img_shape[1]))

        else:
            reshaped_x = np.reshape(x, (-1, img_shape[0] * img_shape[1], img_shape[2]))

        return reshaped_x

    def predict_on_validation(self):
        """ Predict labels by running the model on the validation set.
        
            Returns:
                results: vector
                    A vector containing the predicted labels.                
        """
        x = self.images[DataUse.VALIDATION]
        results = self.get_predictions(x)
        return results

    def predict_on_test(self):
        """ Predict labels by running the model on the test set.
        
            Returns:
                results: vector
                    A vector containing the predicted labels.                
        """
        x = self.images[DataUse.TEST]
        results = self.get_predictions(x)
        return results
    
    def _get_mislabelled(self, x, y, print_report=False, print_detailed_report=False):
        """ Report the number of examples mislabelled by the model.

            Args:
                x:tensor
                    Tensor containing the input data.
                y: tensor
                    Tensor containing the one hot encoded labels.
                print_report: bool
                    If True, prints the percentage of correct and incorrect
                print_detailed_report: bool
                    If True, additionally prints all misclassified examples with the
                    correct and predicted labels.

            Returns:
                results: tuple (numpy arrays)
                    Tuple with two DataFrames (report, incorrect). The first contains
                    number and percentage of correct/incorrect classification. The second,
                    the incorrect examples indices with incorrect and correct labels. 
        
        """
        y1hot = self._ensure1hot(y) 

        x_reshaped = self._reshape_x(x)
        predicted = self.get_predictions(x_reshaped)
        pred_df = pd.DataFrame({"label":np.array(list(map(from1hot,y1hot))), "pred": predicted})
       
        n_predictions = len(pred_df)
        n_correct = sum(pred_df.label == pred_df.pred)
        perc_correct = round(n_correct/n_predictions * 100, 2)
        incorrect = pred_df[pred_df.label != pred_df.pred]
        n_incorrect = len(incorrect)  
        perc_incorrect = round(n_incorrect/n_predictions * 100, 2)
        
        #pred_df.to_csv("predictions.csv")
        report = pd.DataFrame({"correct":[n_correct], "incorrect":[n_incorrect],
                            "%correct":[perc_correct],"%incorrect":[perc_incorrect],
                            "total":[n_predictions]})

        if print_report or print_detailed_report:
            print("=============================================")
            print("Correct classifications: {0} of {1} ({2}%)".format(n_correct, n_predictions, perc_correct))
            print("Incorrect classifications: {0} of {1} ({2})%".format(n_incorrect, n_predictions, perc_incorrect))
            if print_detailed_report:
                print("These were the incorrect classifications:")
                print(incorrect)
            print("=============================================") 
        
        results =(report, incorrect)    
        return results

    def mislabelled_on_validation(self, print_report=False, print_detailed_report=False):
        """ Report the number of examples mislabelled by the trained model on
            the validation set.

            This method wraps around the '_get_mislabelled()' method in the same class.

            Args:
                print_report: bool
                    If True, prints the percentage of correct and incorrect
                print_detailed_report: bool
                    If True, additionally prints all misclassified examples with the
                    correct and predicted labels.
            Returns:
                results: tuple (numpy arrays)
                Tuple with two  DataFrames. The first contains
                number and percentage of correct/incorrect classification. The second,
                the incorrect examples indices with incorrect and correct labels. 
        """
        x = self.images[DataUse.VALIDATION]
        y = self.labels[DataUse.VALIDATION]
        results = self._get_mislabelled(x=x, y=y, print_report=print_report, print_detailed_report=print_detailed_report)
        return results

    def mislabelled_on_test(self, print_report=False, print_detailed_report=False):
        """ Report the number of examples mislabelled by the trained model on
            the test set.

            This method wraps around the '_get_mislabelled()' method in the same class.

            Args:
                print_report: bool
                    If True, prints the percentage of correct and incorrect
                print_detailed_report: bool
                    If True, additionally prints all misclassified examples with the
                    correct and predicted labels.
            Returns:
                results: tuple (numpy arrays)
                Tuple with two  DataFrames. The first contains
                number and percentage of correct/incorrect classification. The second,
                the incorrect examples indices with incorrect and correct labels. 
        """
        x = self.images[DataUse.TEST]
        y = self.labels[DataUse.TEST]
        results = self._get_mislabelled(x=x,y=y, print_report=print_report, print_detailed_report=print_detailed_report)
        return results

    def accuracy_on_train(self):
        """ Report the model accuracy on the training set

            This method wraps around '_check_accuracy()' in the same class.

            Returns:
                results: float
                    The accuracy on the training set.
        """
        x = self.images[DataUse.TRAINING]
        y = self.labels[DataUse.TRAINING]
        results = self._check_accuracy(x,y)
        return results

    def accuracy_on_validation(self):
        """Report the model accuracy on the validation set

            This method wraps around '_check_accuracy()' in the same class.

            Returns:
                results: float
                    The accuracy on the validation set.
        """
        x = self.images[DataUse.VALIDATION]
        y = self.labels[DataUse.VALIDATION]
        results = self._check_accuracy(x, y)
        return results

    def accuracy_on_test(self):
        """Report the model accuracy on the test set

            This method wraps around '_check_accuracy()' in the same class.

            Returns:
                results: float
                    The accuracy on the test set.
        """
        x = self.images[DataUse.TEST]
        y = self.labels[DataUse.TEST]
        results = self._check_accuracy(x,y)
        return results

    def _split(self, x, y=None):
        """ Splits the data into chunks 

            Useful when apply tensorflow operations to large amounts of data

            Args:
                x: tensor
                    Tensor containing the input data
                y: tensor
                    Tensor containing the labels

            Returns:
                x_split: list
                    Input data split up into chunks
                y_split: list
                    Labels split up into chunks
        """

        x_split, y_split = list(), list()

        if x.shape[0] == 0:
            return x_split, y_split

        N = self.max_frames

        imax = int(np.ceil(x.shape[0] / N))
        for i in range(imax):
            i1 = int(N * i)
            i2 = int(min(i1 + N, x.shape[0]))
            xi = x[i1:i2]
            x_split.append(xi)
            if y is not None:
                yi = y[i1:i2]
                y_split.append(yi)

        return x_split, y_split




# wavfile.py (Enhanced)
# Date: 20180430_2335 Joseph Basquin
#
# URL: https://gist.github.com/josephernest/3f22c5ed5dabf1815f16efa8fa53d476
# Source: scipy/io/wavfile.py
#
# Added:
# * read: also returns bitrate, cue markers + cue marker labels (sorted), loops, pitch
#         See https://web.archive.org/web/20141226210234/http://www.sonicspot.com/guide/wavefiles.html#labl
# * read: 24 bit & 32 bit IEEE files support (inspired from wavio_weckesser.py from Warren Weckesser)
# * read: added normalized (default False) that returns everything as float in [-1, 1]
# * read: added forcestereo that returns a 2-dimensional array even if input is mono
#
# * write: can write cue markers, cue marker labels, loops, pitch
# * write: 24 bit support
# * write: can write from a float normalized in [-1, 1] 
# * write: 20180430_2335: bug fixed when size of data chunk is odd (previously, metadata could become unreadable because of this)
#
# * removed RIFX support (big-endian) (never seen one in 10+ years of audio production/audio programming), only RIFF (little-endian) are supported
# * removed read(..., mmap)
#
#
# Test:
# ..\wav\____wavfile_demo.py


"""
Module to read / write wav files using numpy arrays

Functions
---------
`read`: Return the sample rate (in samples/sec) and data from a WAV file.

`write`: Write a numpy array as a WAV file.

"""
from __future__ import division, print_function, absolute_import

import numpy
import struct
import warnings
import collections
from operator import itemgetter

class WavFileWarning(UserWarning):
    pass
    
_ieee = False
    
# assumes file pointer is immediately
#  after the 'fmt ' id
def _read_fmt_chunk(fid):
    res = struct.unpack('<ihHIIHH',fid.read(20))
    size, comp, noc, rate, sbytes, ba, bits = res
    if (comp != 1 or size > 16):
        if (comp == 3):
          global _ieee
          _ieee = True
          #warnings.warn("IEEE format not supported", WavFileWarning)        
        else: 
          warnings.warn("Unfamiliar format bytes", WavFileWarning)
        if (size>16):
            fid.read(size-16)
    return size, comp, noc, rate, sbytes, ba, bits

# assumes file pointer is immediately
#   after the 'data' id
def _read_data_chunk(fid, noc, bits, normalized=False):
    size = struct.unpack('<i',fid.read(4))[0]

    if bits == 8 or bits == 24:
        dtype = 'u1'
        bytes = 1
    else:
        bytes = bits//8
        dtype = '<i%d' % bytes
        
    if bits == 32 and _ieee:
       dtype = 'float32'
        
    data = numpy.fromfile(fid, dtype=dtype, count=size//bytes)
    
    if bits == 24:
        a = numpy.empty((len(data) // 3, 4), dtype='u1')
        a[:, :3] = data.reshape((-1, 3))
        a[:, 3:] = (a[:, 3 - 1:3] >> 7) * 255
        data = a.view('<i4').reshape(a.shape[:-1])
    
    if noc > 1:
        data = data.reshape(-1,noc)
        
    if bool(size & 1):     # if odd number of bytes, move 1 byte further (data chunk is word-aligned)
      fid.seek(1,1)    

    if normalized:
        if bits == 8 or bits == 16 or bits == 24: 
            normfactor = 2 ** (bits-1)
        data = numpy.float32(data) * 1.0 / normfactor

    return data

def _skip_unknown_chunk(fid):
    data = fid.read(4)
    size = struct.unpack('<i', data)[0]
    if bool(size & 1):     # if odd number of bytes, move 1 byte further (data chunk is word-aligned)
      size += 1 
    fid.seek(size, 1)

def _read_riff_chunk(fid):
    str1 = fid.read(4)
    if str1 != b'RIFF':
        raise ValueError("Not a WAV file.")
    fsize = struct.unpack('<I', fid.read(4))[0] + 8
    str2 = fid.read(4)
    if (str2 != b'WAVE'):
        raise ValueError("Not a WAV file.")
    return fsize


def read(file, readmarkers=False, readmarkerlabels=False, readmarkerslist=False, readloops=False, readpitch=False, normalized=False, forcestereo=False):
    """
    Return the sample rate (in samples/sec) and data from a WAV file

    Parameters
    ----------
    file : file
        Input wav file.

    Returns
    -------
    rate : int
        Sample rate of wav file
    data : numpy array
        Data read from wav file

    Notes
    -----

    * The file can be an open file or a filename.

    * The returned sample rate is a Python integer
    * The data is returned as a numpy array with a
      data-type determined from the file.

    """
    if hasattr(file,'read'):
        fid = file
    else:
        fid = open(file, 'rb')

    fsize = _read_riff_chunk(fid)
    noc = 1
    bits = 8
    #_cue = []
    #_cuelabels = []
    _markersdict = collections.defaultdict(lambda: {'position': -1, 'label': ''})
    loops = []
    pitch = 0.0
    while (fid.tell() < fsize):
        # read the next chunk
        chunk_id = fid.read(4)
        if chunk_id == b'fmt ':
            size, comp, noc, rate, sbytes, ba, bits = _read_fmt_chunk(fid)
        elif chunk_id == b'data':
            data = _read_data_chunk(fid, noc, bits, normalized)
        elif chunk_id == b'cue ':
            str1 = fid.read(8)
            size, numcue = struct.unpack('<ii',str1)
            for c in range(numcue):
                str1 = fid.read(24)
                id, position, datachunkid, chunkstart, blockstart, sampleoffset = struct.unpack('<iiiiii', str1)
                #_cue.append(position)
                _markersdict[id]['position'] = position                    # needed to match labels and markers

        elif chunk_id == b'LIST':
            str1 = fid.read(8)
            size, type = struct.unpack('<ii', str1)
        elif chunk_id in [b'ICRD', b'IENG', b'ISFT', b'ISTJ']:    # see http://www.pjb.com.au/midi/sfspec21.html#i5
            _skip_unknown_chunk(fid)
        elif chunk_id == b'labl':
            str1 = fid.read(8)
            size, id = struct.unpack('<ii',str1)
            size = size + (size % 2)                              # the size should be even, see WAV specfication, e.g. 16=>16, 23=>24
            label = fid.read(size-4).rstrip('\x00')               # remove the trailing null characters
            #_cuelabels.append(label)
            _markersdict[id]['label'] = label                           # needed to match labels and markers

        elif chunk_id == b'smpl':
            str1 = fid.read(40)
            size, manuf, prod, sampleperiod, midiunitynote, midipitchfraction, smptefmt, smpteoffs, numsampleloops, samplerdata = struct.unpack('<iiiiiIiiii', str1)
            cents = midipitchfraction * 1./(2**32-1)
            pitch = 440. * 2 ** ((midiunitynote + cents - 69.)/12)
            for i in range(numsampleloops):
                str1 = fid.read(24)
                cuepointid, type, start, end, fraction, playcount = struct.unpack('<iiiiii', str1) 
                loops.append([start, end])
        else:
            warnings.warn("Chunk " + str(chunk_id, 'utf-8') + " skipped", WavFileWarning)
            _skip_unknown_chunk(fid)
    fid.close()

    if data.ndim == 1 and forcestereo:
        data = numpy.column_stack((data, data))

    _markerslist = sorted([_markersdict[l] for l in _markersdict], key=lambda k: k['position'])  # sort by position
    _cue = [m['position'] for m in _markerslist]
    _cuelabels = [m['label'] for m in _markerslist]
    
    return (rate, data, bits, ) \
        + ((_cue,) if readmarkers else ()) \
        + ((_cuelabels,) if readmarkerlabels else ()) \
        + ((_markerslist,) if readmarkerslist else ()) \
        + ((loops,) if readloops else ()) \
        + ((pitch,) if readpitch else ())

  

def write(filename, rate, data, bitrate=None, markers=None, loops=None, pitch=None, normalized=False):
    """
    Write a numpy array as a WAV file

    Parameters
    ----------
    filename : file
        The name of the file to write (will be over-written).
    rate : int
        The sample rate (in samples/sec).
    data : ndarray
        A 1-D or 2-D numpy array of integer data-type.

    Notes
    -----
    * Writes a simple uncompressed WAV file.
    * The bits-per-sample will be determined by the data-type.
    * To write multiple-channels, use a 2-D array of shape
      (Nsamples, Nchannels).

    """

    # normalization and 24-bit handling
    if bitrate == 24:   # special handling of 24 bit wav, because there is no numpy.int24...
        if normalized:
            data[data > 1.0] = 1.0
            data[data < -1.0] = -1.0
            a32 = numpy.asarray(data * (2 ** 23 - 1), dtype=numpy.int32)
        else:
            a32 = numpy.asarray(data, dtype=numpy.int32)
        if a32.ndim == 1:               
            a32.shape = a32.shape + (1,)  # Convert to a 2D array with a single column.
        a8 = (a32.reshape(a32.shape + (1,)) >> numpy.array([0, 8, 16])) & 255  # By shifting first 0 bits, then 8, then 16, the resulting output is 24 bit little-endian.
        data = a8.astype(numpy.uint8)
    else:
        if normalized:   # default to 32 bit int
            data[data > 1.0] = 1.0
            data[data < -1.0] = -1.0
            data = numpy.asarray(data * (2 ** 31 - 1), dtype=numpy.int32)

    fid = open(filename, 'wb')
    fid.write(b'RIFF')
    fid.write(b'\x00\x00\x00\x00')
    fid.write(b'WAVE')

    # fmt chunk
    fid.write(b'fmt ')
    if data.ndim == 1:
        noc = 1
    else:
        noc = data.shape[1]
    bits = data.dtype.itemsize * 8 if bitrate != 24 else 24
    sbytes = rate * (bits // 8) * noc
    ba = noc * (bits // 8)
    fid.write(struct.pack('<ihHIIHH', 16, 1, noc, rate, sbytes, ba, bits))

    # cue chunk
    if markers:    # != None and != []
        if isinstance(markers[0], dict):       # then we have [{'position': 100, 'label': 'marker1'}, ...]
            labels = [m['label'] for m in markers]
            markers = [m['position'] for m in markers]
        else:
            labels = ['' for m in markers]

        fid.write(b'cue ')
        size = 4 + len(markers) * 24
        fid.write(struct.pack('<ii', size, len(markers)))
        for i, c in enumerate(markers):
            s = struct.pack('<iiiiii', i + 1, c, 1635017060, 0, 0, c)           # 1635017060 is struct.unpack('<i',b'data')
            fid.write(s)

        lbls = ''
        for i, lbl in enumerate(labels):
            lbls += b'labl'
            label = lbl + ('\x00' if len(lbl) % 2 == 1 else '\x00\x00')
            size = len(lbl) + 1 + 4          # because \x00
            lbls += struct.pack('<ii', size, i + 1)
            lbls += label

        fid.write(b'LIST')
        size = len(lbls) + 4 
        fid.write(struct.pack('<i', size))                         
        fid.write(b'adtl')                                                      # https://web.archive.org/web/20141226210234/http://www.sonicspot.com/guide/wavefiles.html#list
        fid.write(lbls)        

    # smpl chunk
    if loops or pitch:
      if not loops:
        loops = []
      if pitch:
        midiunitynote = 12 * numpy.log2(pitch * 1.0 / 440.0) + 69
        midipitchfraction = int((midiunitynote - int(midiunitynote)) * (2**32-1))
        midiunitynote = int(midiunitynote)
        #print(midipitchfraction, midiunitynote)
      else:
        midiunitynote = 0
        midipitchfraction = 0
      fid.write(b'smpl')
      size = 36 + len(loops) * 24
      sampleperiod = int(1000000000.0 / rate)

      fid.write(struct.pack('<iiiiiIiiii', size, 0, 0, sampleperiod, midiunitynote, midipitchfraction, 0, 0, len(loops), 0))
      for i, loop in enumerate(loops):
        fid.write(struct.pack('<iiiiii', 0, 0, loop[0], loop[1], 0, 0))

    # data chunks
    fid.write(b'data')
    fid.write(struct.pack('<i', data.nbytes))
    import sys
    if data.dtype.byteorder == '>' or (data.dtype.byteorder == '=' and sys.byteorder == 'big'):
        data = data.byteswap()

    data.tofile(fid)

    if data.nbytes % 2 == 1: # add an extra padding byte if data.nbytes is odd: https://web.archive.org/web/20141226210234/http://www.sonicspot.com/guide/wavefiles.html#data
        fid.write('\x00')

    # Determine file size and place it in correct
    #  position at start of the file.
    size = fid.tell()
    fid.seek(4)
    fid.write(struct.pack('<i', size-8))
    fid.close()



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the data_feeding module within the ketos library
"""

import os
import pytest
import numpy as np
import pandas as pd
from tables import open_file
from ketos.data_handling.database_interface import open_table
from ketos.data_handling.data_feeding import ActiveLearningBatchGenerator, BatchGenerator
from ketos.neural_networks.neural_networks import class_confidences, predictions

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')


@pytest.mark.test_BatchGenerator
def test_one_batch():
    """ Test if one batch has the expected shape and contents
    """
    h5 = open_file(os.path.join(path_to_assets, "15x_same_spec.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/species1")

    five_specs = train_data[:5]['data']
    five_boxes = train_data[:5]['boxes']

    train_generator = BatchGenerator(hdf5_table=train_data, batch_size=5, return_batch_ids=True) #create a batch generator 
    ids, X, Y = next(train_generator)
    assert ids == [0,1,2,3,4]
    assert X.shape == (5, 2413, 201)
    np.testing.assert_array_equal(X, five_specs)
    assert Y.shape == (5,)
    np.testing.assert_array_equal(Y, five_boxes)

    h5.close()


@pytest.mark.test_BatchGenerator
def test_labels_as_Y():
    """ Test if batch generator returns labels instead of boxes
    """
    h5 = open_file(os.path.join(path_to_assets, "15x_same_spec.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/species1")
    
    five_labels = train_data[:5]['labels']

    train_generator = BatchGenerator(hdf5_table=train_data, y_field='labels', batch_size=5, return_batch_ids=False) #create a batch generator 
    _, Y = next(train_generator)
    np.testing.assert_array_equal(Y, five_labels)

    h5.close()
    

@pytest.mark.test_BatchGenerator
def test_batch_sequence_same_as_db():
    """ Test if batches are generated with instances in the same order as they appear in the database
    """
    h5 = open_file(os.path.join(path_to_assets, "15x_same_spec.h5"), 'r') #create the database handle  
    train_data = open_table(h5, "/train/species1")

    ids_in_db = train_data[:]['id']
    train_generator = BatchGenerator(hdf5_table=train_data, x_field='id', batch_size=3, return_batch_ids=True) #create a batch generator 

    for i in range(3):
        ids, X, _ = next(train_generator)
        np.testing.assert_array_equal(X, ids_in_db[i*3: i*3+3])
        assert ids == list(range(i*3, i*3+3))
    
    h5.close()


@pytest.mark.test_BatchGenerator
def test_last_batch():
    """ Test if last batch has the expected number of instances
    """
    h5 = open_file(os.path.join(path_to_assets, "15x_same_spec.h5"), 'r') #create the database handle  
    train_data = open_table(h5, "/train/species1")

    ids_in_db = train_data[:]['id']
    train_generator = BatchGenerator(hdf5_table=train_data, batch_size=6, return_batch_ids=True) #create a batch generator 
    #First batch
    ids, X, _ = next(train_generator)
    assert ids == [0,1,2,3,4,5]
    assert X.shape == (6, 2413, 201)
    #Second batch
    ids, X, _ = next(train_generator)
    assert ids == [6,7,8,9,10,11]
    assert X.shape == (6, 2413, 201)
    #last batch
    ids, X, _ = next(train_generator)
    assert ids == [12,13,14]
    assert X.shape == (3, 2413, 201)

    h5.close()

@pytest.mark.test_BatchGenerator
def test_use_only_subset_of_data():
    """ Test that only the indices specified are used
    """
    h5 = open_file(os.path.join(path_to_assets, "15x_same_spec.h5"), 'r') #create the database handle  
    train_data = open_table(h5, "/train/species1")

    train_generator = BatchGenerator(hdf5_table=train_data, indices=[1,3,5,7,9,11,13], batch_size=4, return_batch_ids=True) #create a batch generator 
    #First batch
    ids, X, _ = next(train_generator)
    assert ids == [1,3,5,7]
    #Second batch
    ids, X, _ = next(train_generator)
    assert ids == [9,11,13]

    h5.close()

@pytest.mark.test_BatchGenerator
def test_multiple_epochs():
    """ Test if batches are as expected after the first epoch
    """
    h5 = open_file(os.path.join(path_to_assets, "15x_same_spec.h5"), 'r') #create the database handle  
    train_data = open_table(h5, "/train/species1")

    ids_in_db = train_data[:]['id']
    train_generator = BatchGenerator(hdf5_table=train_data, batch_size=6, return_batch_ids=True) #create a batch generator 
    #Epoch 0, batch 0
    ids, X, _ = next(train_generator)
    assert ids == [0,1,2,3,4,5]
    assert X.shape == (6, 2413, 201)
    #Epoch 0, batch 1
    ids, X, _ = next(train_generator)
    assert ids == [6,7,8,9,10,11]
    assert X.shape == (6, 2413, 201)
    #Epoch 0 batch2
    ids, X, _ = next(train_generator)
    assert ids == [12,13,14]
    assert X.shape == (3, 2413, 201)

    #Epoch 1, batch 0
    ids, X, _ = next(train_generator)
    assert ids == [0,1,2,3,4,5]
    assert X.shape == (6, 2413, 201)

    h5.close()

@pytest.mark.test_BatchGenerator
def test_load_from_memory():
    """ Test if batch generator can work with data loaded from memory
    """
    x = np.ones(shape=(15,32,16))
    y = np.zeros(shape=(15))

    generator = BatchGenerator(x=x, y=y, batch_size=6, return_batch_ids=True) #create a batch generator 

    #Epoch 0, batch 0
    ids, X, _ = next(generator)
    assert ids == [0,1,2,3,4,5]
    assert X.shape == (6, 32, 16)
    #Epoch 0, batch 1
    ids, X, _ = next(generator)
    assert ids == [6,7,8,9,10,11]
    assert X.shape == (6, 32, 16)
    #Epoch 0 batch2
    ids, X, _ = next(generator)
    assert ids == [12,13,14]
    assert X.shape == (3, 32, 16)

    #Epoch 1, batch 0
    ids, X, _ = next(generator)
    assert ids == [0,1,2,3,4,5]
    assert X.shape == (6, 32, 16)

@pytest.mark.test_BatchGenerator
def test_shuffle():
    """Test shuffle argument.
        Instances should be shuffled before divided into batches, but the order should be consistent across epochs if
        'refresh_on_epoch_end' is False.
    """
    h5 = open_file(os.path.join(path_to_assets, "15x_same_spec.h5"), 'r') #create the database handle  
    train_data = open_table(h5, "/train/species1")


    ids_in_db = train_data[:]['id']
    np.random.seed(100)
    
    train_generator = BatchGenerator(hdf5_table=train_data, batch_size=6, return_batch_ids=True, shuffle=True) #create a batch generator 

    for epoch in range(5):
        #batch 0
        ids, X, _ = next(train_generator)
        assert ids == [9, 1, 12, 13, 6, 10]
        assert X.shape == (6, 2413, 201)
        #batch 1
        ids, X, _ = next(train_generator)
        assert ids == [5, 2, 4, 0, 11, 7]
        assert X.shape == (6, 2413, 201)
        #batch 2
        ids, X, _ = next(train_generator)
        assert ids ==  [3, 14, 8]
        assert X.shape == (3, 2413, 201)
    
    h5.close()


@pytest.mark.test_BatchGenerator
def test_refresh_on_epoch_end():
    """ Test if batches are generated with randomly selected instances for each epoch
    """
    h5 = open_file(os.path.join(path_to_assets, "15x_same_spec.h5"), 'r') #create the database handle  
    train_data = open_table(h5, "/train/species1")

    ids_in_db = train_data[:]['id']
    np.random.seed(100)
    
    train_generator = BatchGenerator(hdf5_table=train_data, batch_size=6, return_batch_ids=True, shuffle=True, refresh_on_epoch_end=True) #create a batch generator 

    expected_ids = {'epoch_1':([9, 1, 12, 13, 6, 10],[5, 2, 4, 0, 11, 7],[3, 14, 8]),
                     'epoch_2': ([0, 2, 1, 14, 10, 3],[13, 12, 5, 8, 11, 7],[6, 4, 9]),    
                     'epoch_3': ([7, 13, 1, 0, 11, 9],[5, 8, 2, 12, 4, 6],[10, 14, 3])}
                     
    for epoch in ['epoch_1', 'epoch_2', 'epoch_3']:
        #batch 0
        ids, X, _ = next(train_generator)
        print(epoch)
        assert ids == expected_ids[epoch][0]
        #batch 1
        ids, X, _ = next(train_generator)
        assert ids == expected_ids[epoch][1]
        #batch 2
        ids, X, _ = next(train_generator)
        assert ids == expected_ids[epoch][2]
    
    h5.close()

@pytest.mark.test_BatchGenerator
def test_instance_function():
    """ Test if the function passed as 'instance_function' is applied to the batch
    """
    h5 = open_file(os.path.join(path_to_assets, "15x_same_spec.h5"), 'r') # create the database handle  
    train_data = open_table(h5, "/train/species1")

    def apply_to_batch(X,Y):
        X = np.mean(X, axis=(1,2))
        return (X, Y)

    train_generator = BatchGenerator(hdf5_table=train_data, batch_size=5, return_batch_ids=True, instance_function=apply_to_batch) #create a batch generator 
    _, X, Y = next(train_generator)
    assert X.shape == (5,)
    assert X[0] == pytest.approx(7694.1147, 0.1)
    assert Y.shape == (5,)
    
    h5.close()


@pytest.mark.test_ActiveLearningBatchGenerator
def test_active_learning_batch_generator_max_keep_zero():
    """ Test can start first training session
    """
    h5 = open_file(os.path.join(path_to_assets, "15x_same_spec.h5"), 'r')  
    data = open_table(h5, "/train/species1")

    specs = data[:]['data']
    labels = data[:]['labels']

    a = ActiveLearningBatchGenerator(table=data, session_size=6, batch_size=2, return_indices=True)

    # get 1st batch generator
    generator = next(a)

    # get 1st batch
    ids, X, Y = next(generator)
    assert ids == [0,1]
    assert X.shape == (2, 2413, 201)
    np.testing.assert_array_equal(X, specs[:2])
    assert len(Y) == 2
    assert Y == [1,1]

    # get 2nd batch
    ids, X, Y = next(generator)
    assert ids == [2,3]
    assert X.shape == (2, 2413, 201)
    np.testing.assert_array_equal(X, specs[2:4])
    assert len(Y) == 2
    assert Y == [1,1]

    # get 3rd and 4th batch
    ids, _, _ = next(generator) 
    assert ids == [4,5]
    ids, _, _ = next(generator) 
    assert ids == [0,1]

    # get 2nd batch generator
    generator = next(a)

    # get batches
    ids, _, _ = next(generator) 
    assert ids == [6,7]
    ids, _, _ = next(generator) 
    assert ids == [8,9]
    ids, _, _ = next(generator) 
    assert ids == [10,11]
    ids, _, _ = next(generator) 
    assert ids == [6,7]

    # get 3rd batch generator
    generator = next(a)

    # get batches
    ids, _, _ = next(generator) 
    assert ids == [12,13]
    ids, _, _ = next(generator) 
    assert ids == [14,0]
    ids, _, _ = next(generator) 
    assert ids == [1,2]

    h5.close()


@pytest.mark.test_ActiveLearningBatchGenerator
def test_active_learning_batch_generator_max_keep_nonzero():
    """ Test can start first training session
    """
    h5 = open_file(os.path.join(path_to_assets, "15x_same_spec.h5"), 'r')  
    data = open_table(h5, "/train/species1")

    a = ActiveLearningBatchGenerator(table=data, session_size=7, batch_size=3, max_keep=0.3, return_indices=True)

    generator = next(a)

    ids, _, _ = next(generator)
    assert ids == [0,1,2]
    ids, _, _ = next(generator)
    assert ids == [3,4,5]
    ids, _, _ = next(generator)
    assert ids == [6]

    generator = next(a)

    ids, _, Y = next(generator)
    assert ids == [7,8,9]
    assert Y == [1,1,1]
    a.update_performance(indices=[7,8,9], predictions=[1,1,1], confidences=[1.0,1.0,1.0])

    ids, _, Y = next(generator)
    assert ids == [10,11,12]
    assert Y == [1,1,1]
    a.update_performance(indices=[10,11,12], predictions=[1,1,1], confidences=[1.0,1.0,1.0])

    ids, _, Y = next(generator)
    assert ids == [13]
    assert Y == 1
    a.update_performance(indices=[13], predictions=[0], confidences=[1.0])

    generator = next(a)

    ids1, _, Y = next(generator)
    assert Y == [1,1,1]
    a.update_performance(indices=ids1, predictions=[1,0,0])

    ids2, _, Y = next(generator)
    assert Y == [1,1,1]
    a.update_performance(indices=ids2, predictions=[0,0,0])

    ids3, _, Y = next(generator)
    assert Y == 1
    a.update_performance(indices=ids3, predictions=[1])

    ids = np.concatenate((ids1, ids2, ids3))
    assert 13 in ids

    wrong_ids = [ids1[1], ids1[2], ids2[0], ids2[1], ids2[2]]

    generator = next(a)

    ids1, _, _ = next(generator)
    ids2, _, _ = next(generator)
    ids3, _, _ = next(generator)
    ids = np.concatenate((ids1, ids2, ids3))

    # check how many of the wrong predictiosn were kept
    # since max_keep = 0.3 and session_size = 7, it should be int(0.3*7) = 2
    num_keep = 0
    for id in wrong_ids:
        num_keep += (id in ids)

    assert num_keep == 2

    h5.close()


@pytest.mark.test_ActiveLearningBatchGenerator
def test_active_learning_batch_generator_load_from_memory():
    """ Test can start first and second training session
    """
    x = np.ones(shape=(15,32,16))
    y = np.zeros(shape=(15))

    a = ActiveLearningBatchGenerator(x=x, y=y, session_size=7, batch_size=3, return_indices=True)

    generator = next(a)

    ids, _, _ = next(generator)
    assert ids == [0,1,2]
    ids, _, _ = next(generator)
    assert ids == [3,4,5]
    ids, _, _ = next(generator)
    assert ids == [6]

    generator = next(a)

    ids, _, _ = next(generator)
    assert ids == [7,8,9]


@pytest.mark.test_ActiveLearningBatchGenerator
def test_active_learning_batch_generator_splits_into_training_and_validation():
    """ Test can split into training and validation data
    """
    h5 = open_file(os.path.join(path_to_assets, "15x_same_spec.h5"), 'r')  
    data = open_table(h5, "/train/species1")

    specs = data[:]['data']
    labels = data[:]['labels']

    a = ActiveLearningBatchGenerator(table=data, session_size=6, batch_size=2, num_labels=2,\
        return_indices=True, seed=1, val_frac=0.2, convert_to_one_hot=True)

    assert len(a.val_indices) == 3

    generator = next(a)
    indices = np.concatenate((generator.entry_indices, a.val_indices))

    generator = next(a)
    indices = np.concatenate((generator.entry_indices, indices))

    indices = np.sort(indices)
    indices = np.unique(indices)

    assert np.all(indices == np.arange(15))

    idx, X, Y = a.get_validation_data()
    assert np.all(idx == a.val_indices)

    assert Y.shape[0] == 3
    assert Y.shape[1] == 2

    h5.close()
